import math, pygame
from pygame.locals import *

########################
# Initialize variables #
########################

n = 8   # Number of points

width = height = 500
xCenter = width/2
yCenter = height/2
fps = 30

bgColor = (255,255,255)
axisColor = (128,128,128)
lineColor = (0,0,0)
scale = 200

rConst = math.pi/15     # Radians of rotation per frame

origin = (.5,.5,.5)    # Translate the point of rotation
x = "x"
y = "y"
z = "z"

# Initialize 2-dimensional lists. Point number followed by component.
# Ex: [0, 1, 2, 3, ... , n][x,y,z]
# 'point' has the true coordinates of the points
# 'pointPlot' has the projected coordinates
point = [[0 for coord in range(3)] for pointNumber in range(n)]
pointPlot = [[0 for coord in range(3)] for pointNumber in range(n)]

########################
# Function definitions #
########################

def setCoord(pointNum,x,y,z):
    """Set x, y, z coords for a specified point"""
    point[pointNum][0] = x
    point[pointNum][1] = y
    point[pointNum][2] = z
    setPlotCoord(pointNum,x-origin[0],y-origin[1],z-origin[2])

def setPlotCoord(pointNum,xPlot,yPlot,zPlot):
    """Set values for projected coordinates
        (To be used in rotation functions)"""
    pointPlot[pointNum][0] = xPlot
    pointPlot[pointNum][1] = yPlot
    pointPlot[pointNum][2] = zPlot

def getCoord(pointNum):
    """Return true x, y, and z coords"""
    return point[pointNum][0],point[pointNum][1],point[pointNum][2]

def getPlotCoord(pointNum):
    """Return projected x, y, and z coords"""
    return pointPlot[pointNum][0],pointPlot[pointNum][1],pointPlot[pointNum][2]

def xRotate(pointNum,posOrNeg):
    """Rotate a point about the x-axis
        posOrNeg should be either 1 or -1"""
    x,y,z = getPlotCoord(pointNum)
    angle = rConst*posOrNeg
    xPlot = x
    yPlot = y*math.cos(angle) - z*math.sin(angle)
    zPlot = z*math.cos(angle) + y*math.sin(angle)
    setPlotCoord(pointNum,xPlot,yPlot,zPlot)

def yRotate(pointNum,posOrNeg):
    """Rotate a point about the y-axis
        posOrNeg should be either 1 or -1"""
    x,y,z = getPlotCoord(pointNum)
    angle = rConst*posOrNeg
    xPlot = math.cos(angle)*x - z*math.sin(angle)
    yPlot = y
    zPlot = math.sin(angle)*x + z*math.cos(angle)
    setPlotCoord(pointNum,xPlot,yPlot,zPlot)

def zRotate(pointNum,posOrNeg):
    """Rotate a point about the z-axis
        posOrNeg should be either 1 or -1"""
    x,y,z = getPlotCoord(pointNum)
    angle = rConst*posOrNeg
    xPlot = math.cos(angle)*x - y*math.sin(angle)
    yPlot = math.sin(angle)*x + y*math.cos(angle)
    zPlot = z
    setPlotCoord(pointNum,xPlot,yPlot,zPlot)

def rotatePoints(component,posOrNeg):
    """Rotate all points about a specified axis"""
    for pointNum in range(n):
        if component == "x":
            xRotate(pointNum,posOrNeg)
        if component == "y":
            yRotate(pointNum,posOrNeg)
        if component == "z":
            zRotate(pointNum,posOrNeg)

def xCoord(xValue):
    """Converts x value to an x coord to be displayed on the window"""
    return scale*xValue+xCenter

def yCoord(yValue):
    """Converts y value to a y coord to be displayed on the window"""
    return -scale*yValue+yCenter

def getXY(pointNum):
    """Returns a tuple containing projected x and y coords"""
    return (xCoord(pointPlot[pointNum][0]),yCoord(pointPlot[pointNum][1]))

def refreshScreen():
    """Clears screen and redraws all lines (including axes)"""
    screen.fill(bgColor)
    pygame.draw.aaline(screen,axisColor,(0,yCenter),(width,yCenter),1)
    pygame.draw.aaline(screen,axisColor,(xCenter,0),(xCenter,height),1)
    for contLine in range(len(drawingOrder)):
        for lineSeg in range(len(drawingOrder[contLine])-1):
            pygame.draw.aaline(screen,lineColor,getXY(drawingOrder[contLine][lineSeg]),getXY(drawingOrder[contLine][lineSeg+1]),1)
    pygame.display.flip()

########
# Main #
########

setCoord(0,0,0,0)
setCoord(1,0,0,1)
setCoord(2,0,1,0)
setCoord(3,0,1,1)
setCoord(4,1,0,0)
setCoord(5,1,0,1)
setCoord(6,1,1,0)
setCoord(7,1,1,1)

drawingOrder = [(0,1,3,2,0,4,5,1),(3,7,5),(4,6,7),(2,6)]

# Load a screen for graphical output
pygame.init()
screen = pygame.display.set_mode((width,height))
screen.fill(bgColor)
pygame.display.set_caption('3D transformation matrix test')
refreshScreen()

running = True
while running:
    for event in pygame.event.get():
        if event.type == QUIT:
            running = False
        if event.type == KEYDOWN and event.key == K_ESCAPE:
            running = False

    pygame.time.wait(math.floor(1/fps*1000))
    key = pygame.key.get_pressed()
    if key[K_s]:
        rotatePoints(x,1)
    if key[K_w]:
        rotatePoints(x,-1)
    if key[K_a]:
        rotatePoints(y,1)
    if key[K_d]:
        rotatePoints(y,-1)
    if key[K_q]:
        rotatePoints(z,1)
    if key[K_e]:
        rotatePoints(z,-1)
    refreshScreen()

pygame.display.quit()
