# n-body gravity simulator
# Simulates the effects of gravity in a system of n objects in space.
# by William Ho

import math, pygame
from pygame.locals import *
from datetime import datetime

startTime = datetime.now()

#######################
# Values of constants #
#######################

fileName = "gravityspeedcontest.txt"

# Define variables for graphical output
width = height = 500
lineColor = (255,255,255)

########################
# Function definitions #
########################
    
def forceOverMasses(obj1,obj2):
    """Returns the value of the force between two objects (divided by m1*m2).
       In terms of an equation:   F / (m1*m2) = abs(r_x1 - r_x2) / r^3"""
    # Calculate the inverse of the distance between two points, cubed
    rCubed = ((rTemp[obj2][0]-rTemp[obj1][0])**2        \
              +(rTemp[obj2][1]-rTemp[obj1][1])**2       \
              +(rTemp[obj2][2]-rTemp[obj1][2])**2       \
              )**(-1.5)

    # Calculate the abs value of the numerator (r_x1 - r_x2) with offset
    numerator = [0.0]*3 
    for comp in range(3):   # Set values for x, y, z components of numerator
        numerator[comp] = abs(rTemp[obj1][comp]-rTemp[obj2][comp])
    return numerator[0]*rCubed,numerator[1]*rCubed,numerator[2]*rCubed

def forceOverMassesArray():
    """Stores all the unique values of 'abs(r_x1 - r_x2) / r^3' in an array"""
    for obj1 in range(n):
        for obj2 in range(n):
            if obj1 < obj2:
                # Since the force of 1 on 2 is equal in magnitude
                # to the force of 2 on 1, only one of these is stored
                FOverM[0][obj1][obj2],    \
                FOverM[1][obj1][obj2],    \
                FOverM[2][obj1][obj2] =   \
                forceOverMasses(obj1,obj2)

def acc(obj1, comp):
    """Returns the component of acceleration for the specified object"""
    # Calculate the total force divided by the mass of the specified object
    netFOverM = 0
    for obj2 in range(n):
        # Determine the sign of the force based on relative positions
        if rTemp[obj1][comp] > rTemp[obj2][comp]:
            posOrNeg = -1
        elif rTemp[obj1][comp] < rTemp[obj2][comp]:
            posOrNeg = 1
        elif rTemp[obj1][comp] == rTemp[obj2][comp]:
            posOrNeg = 0

        # Add to the total F/m
        if obj1 > obj2:
            netFOverM += posOrNeg * FOverM[comp][obj2][obj1] * m[obj2]
        elif obj1 < obj2:
            netFOverM += posOrNeg * FOverM[comp][obj1][obj2] * m[obj2]
    return netFOverM

def RK4():
    """Updates values for position and velocity using Runge-Kutta 4"""
    for slopeNum in range(4):
        # offset represents the value that is multiplied by the RKSlope
        if slopeNum == 0:
            offset = 0
        elif slopeNum == 1 or slopeNum == 2:
            offset = 0.5*h
        elif slopeNum == 3:
            offset = h

        for obj in range(n):
            for comp in range(6):
                rTemp[obj][comp] = r[obj][comp]+offset*RKSlope[slopeNum-1][obj][comp]
        
        # Determine the RKSlope for each object
        for obj in range(n):
            forceOverMassesArray()
            for comp in range(3):
                RKSlope[slopeNum][obj][comp] = r[obj][comp+3]+offset*RKSlope[slopeNum-1][obj][comp+3]
                RKSlope[slopeNum][obj][comp+3] = acc(obj,comp)

    # Update all values for position and velocity
    for obj in range (n):       
        for comp in range(6):
            r[obj][comp] += hConst*(RKSlope[0][obj][comp] \
                            +2*RKSlope[1][obj][comp] \
                            +2*RKSlope[2][obj][comp] \
                            +RKSlope[3][obj][comp])

def xCoord(xValue):
    """Converts the value of x into an x value on the graph"""
    return xValue+xCenter

def yCoord(yValue):
    """Converts the value of y into a y value on the graph"""
    return -yValue+yCenter

def KE():
    """Returns the total kinetic energy in the system at the current time"""
    totalKE = 0
    for obj in range(n):
        totalKE += .5*m[obj]*(r[obj][3]**2+r[obj][4]**2+r[obj][5]**2)
    return totalKE

def PE():
    """Returns the total potential energy in the system at the current time"""
    totalPE = 0
    for obj1 in range(n):
        for obj2 in range(n):
            if obj1 < obj2:
                distInv = 1/((r[obj2][0]-r[obj1][0])**2     \
                          +(r[obj2][1]-r[obj1][1])**2       \
                          +(r[obj2][2]-r[obj1][2])**2)**.5
                totalPE -= (m[obj1]*m[obj2])*distInv
    return totalPE

########
# Main #
########

file = open(fileName,"r")
n = int(file.readline())      # Read number of objects
m = [0.0]*n

# r array consists of [obj 1, obj 2, ... , obj n][r_x, r_y, r_z, v_x, v_y, v_z]
rOld = [[0.0 for component in range(6)] for objNumber in range(n)]
r = [[0.0 for component in range(6)] for objNumber in range(n)]
rTemp = [[0.0 for component in range(6)] for objNumber in range(n)]

# Read values from the input file and insert them into the appropriate arrays
for i in range(n):
    m[i] = float(file.readline())               # Read values for masses
    for j in range(2):
        rOrV = j*3      # if 0, this will refer to position; if 1, velocity
        rOld[i][rOrV],rOld[i][rOrV+1],rOld[i][rOrV+2] = file.readline().split(", ")     # Read values for r_x, r_y, r_z, v_x, v_y, v_z
        for component in range(3):
            rOld[i][rOrV+component] = float(rOld[i][rOrV+component])

# Set r array equal to rOld array
for obj in range(n):
    for comp in range(6):
        r[obj][comp] = rTemp[obj][comp] = rOld[obj][comp]

h = float(file.readline())      # Step size
hConst = h/6                    # Constant used in RK4
tFinal = float(file.readline()) # Final time

# Initialize global arrays that will be used in functions
FOverM = [[[0.0 for obj2 in range(n)] for obj1 in range(n)]        \
           for component in range(3)]

RKSlope = [[[0.0 for comp in range(6)] for objNum in range(n)]      \
           for slopeNum in range(4)]

# Load a screen for graphical output
pygame.init()
screen = pygame.display.set_mode((width,height))
xCenter = width/2
yCenter = height/2

EInitial = KE()+PE()    # Compute initial energy

# Plot solution curves to the screen
t = 0
while t < tFinal:
    RK4()   # Update position and velocity for all objects
    for obj in range(n):
        pygame.draw.line(screen,lineColor,((xCoord(rOld[obj][0]),    \
            yCoord(rOld[obj][1]))),(xCoord(r[obj][0]),yCoord(r[obj][1])),1)
    t += h
    for obj in range(n):    # Update values in rOld to current values
        for comp in range(6):
            rOld[obj][comp] = r[obj][comp]

pygame.display.flip()   # Update screen
endTime = datetime.now()

EFinal = KE()+PE()      # Compute final energy

print("Time elapsed:",endTime-startTime)
print("EInitial = ",EInitial)
print("EFinal   = ",EFinal)

running = True
# Program finishes if ESC is pressed or the graphics window is closed.
while running:
    for event in pygame.event.get():
        if event.type == QUIT:
            running = False
        if event.type == KEYDOWN and event.key == K_ESCAPE:
            running = False
pygame.display.quit()
