# n-body gravity simulator
# Simulates the effects of gravity in a system of n objects in space.
# by William Ho

import math, pygame, random
from pygame.locals import *
from datetime import datetime

#######################
# Values of constants #
#######################

t = 0
h = .1
hConst = h/6

# Define variables for graphical output
width = height = 500
bgColor = (255,255,255)
axisColor = (128,128,128)
scale = 2

########################
# Function definitions #
########################

def dInput(string,default):
    """A version of the input() function with a default value if
        nothing is entered"""
    print(string," [",default,"]:",sep="",end=" ")
    inputValue = input()
    if inputValue == "":
        inputValue = default
    else:
        inputValue = inputValue
    return inputValue
    
def forceOverMasses(obj1,obj2):
    """Returns the value of the force between two objects (divided by m1*m2).
       In terms of an equation:   F / (m1*m2) = abs(r_x1 - r_x2) / r^3"""
    # Calculate the inverse of the distance between two points, cubed
    rCubed = ((rTemp[obj2][0]-rTemp[obj1][0])**2        \
              +(rTemp[obj2][1]-rTemp[obj1][1])**2       \
              +(rTemp[obj2][2]-rTemp[obj1][2])**2       \
              +1)**(-1.5)

    # Calculate the abs value of the numerator (r_x1 - r_x2) with offset
    numerator = [0.0]*3 
    for comp in range(3):   # Set values for x, y, z components of numerator
        numerator[comp] = abs(rTemp[obj1][comp]-rTemp[obj2][comp])
    return numerator[0]*rCubed,numerator[1]*rCubed,numerator[2]*rCubed

def forceOverMassesArray():
    """Stores all the unique values of 'abs(r_x1 - r_x2) / r^3' in an array"""
    for obj1 in range(n):
        for obj2 in range(n):
            if obj1 < obj2:
                # Since the force of 1 on 2 is equal in magnitude
                # to the force of 2 on 1, only one of these is stored
                FOverM[0][obj1][obj2],    \
                FOverM[1][obj1][obj2],    \
                FOverM[2][obj1][obj2] =   \
                forceOverMasses(obj1,obj2)

def acc(obj1, comp):
    """Returns the component of acceleration for the specified object"""
    # Calculate the total force divided by the mass of the specified object
    netFOverM = 0
    for obj2 in range(n):
        # Determine the sign of the force based on relative positions
        if rTemp[obj1][comp] > rTemp[obj2][comp]:
            posOrNeg = -1
        elif rTemp[obj1][comp] < rTemp[obj2][comp]:
            posOrNeg = 1
        elif rTemp[obj1][comp] == rTemp[obj2][comp]:
            posOrNeg = 0
        if m[obj1] < 0:
            posOrNeg *= -1

        # Add to the total F/m
        if obj1 > obj2:
            netFOverM += posOrNeg * FOverM[comp][obj2][obj1] * m[obj2]
        elif obj1 < obj2:
            netFOverM += posOrNeg * FOverM[comp][obj1][obj2] * m[obj2]
    return netFOverM

def RK4():
    """Updates values for position and velocity using Runge-Kutta 4"""
    for slopeNum in range(4):
        # offset represents the value that is multiplied by the RKSlope
        if slopeNum == 0:
            offset = 0
        elif slopeNum == 1 or slopeNum == 2:
            offset = 0.5*h
        elif slopeNum == 3:
            offset = h

        for obj in range(n):
            for comp in range(6):
                rTemp[obj][comp] = r[obj][comp]+offset*RKSlope[slopeNum-1][obj][comp]
        
        # Determine the RKSlope for each object
        for obj in range(n):
            forceOverMassesArray()
            for comp in range(3):
                RKSlope[slopeNum][obj][comp] = r[obj][comp+3]+offset*RKSlope[slopeNum-1][obj][comp+3]
                RKSlope[slopeNum][obj][comp+3] = acc(obj,comp)

    # Update all values for position and velocity
    for obj in range (n):
        for comp in range(6):
            r[obj][comp] += hConst*(RKSlope[0][obj][comp] \
                            +2*RKSlope[1][obj][comp] \
                            +2*RKSlope[2][obj][comp] \
                            +RKSlope[3][obj][comp])


def xCoord(xValue):
    """Converts the value of x into an x value on the graph"""
    return scale*xValue+xCenter

def yCoord(yValue):
    """Converts the value of y into a y value on the graph"""
    return -scale*yValue+yCenter

def KE():
    """Returns the total kinetic energy in the system at the current time"""
    totalKE = 0
    for obj in range(n):
        totalKE += .5*m[obj]*(r[obj][3]**2+r[obj][4]**2+r[obj][5]**2)
    return totalKE

def PE():
    """Returns the total potential energy in the system at the current time"""
    totalPE = 0
    for obj1 in range(n):
        for obj2 in range(n):
            if obj1 < obj2:
                distInv = 1/((r[obj2][0]-r[obj1][0])**2     \
                          +(r[obj2][1]-r[obj1][1])**2       \
                          +(r[obj2][2]-r[obj1][2])**2)**.5
                totalPE -= (m[obj1]*m[obj2])*distInv
    return totalPE

########
# Main #
########

n = 30
m = [1]*n
lineColor = [0]*n

for obj in range(n):
    lineColor[obj] = (random.randint(0,255),random.randint(0,255),random.randint(0,255))

# r array consists of [obj 1, obj 2, ... , obj n][r_x, r_y, r_z, v_x, v_y, v_z]
rOld = [[0.0 for component in range(6)] for objNumber in range(n)]
r = [[0.0 for component in range(6)] for objNumber in range(n)]
rTemp = [[0.0 for component in range(6)] for objNumber in range(n)]

# Randomize values
for i in range(n):
    for j in range(3):
        rOld[i][j] = random.uniform(-100,100)
    for j in range(3,6):
        rOld[i][j] = random.uniform(-.25,.25)

# Set r array equal to rOld array
for obj in range(n):
    for comp in range(6):
        r[obj][comp] = rTemp[obj][comp] = rOld[obj][comp]

# Initialize global arrays that will be used in functions
FOverM = [[[0.0 for obj2 in range(n)] for obj1 in range(n)]        \
           for component in range(3)]

RKSlope = [[[0.0 for comp in range(6)] for objNum in range(n)]      \
           for slopeNum in range(4)]

# Load a screen for graphical output
pygame.init()
screen = pygame.display.set_mode((width,height))
screen.fill(bgColor)
xCenter = width/2
yCenter = height/2
pygame.draw.line(screen,axisColor,(yCenter,0),(yCenter,width),1)
pygame.draw.line(screen,axisColor,(0,xCenter),(height,xCenter),1)

EInitial = KE()+PE()    # Compute initial energy

startTime = datetime.now()

# Plot solution curves to the screen
t = 0
plotting = True
while plotting == True:
    RK4()   # Update position and velocity for all objects
    for obj in range(n):
        pygame.draw.aaline(screen,lineColor[obj],
            (xCoord(rOld[obj][0]), yCoord(rOld[obj][1])), \
            (xCoord(r[obj][0]),yCoord(r[obj][1])),1)
    pygame.display.flip()
    t += h

    for obj in range(n):    # Update values in rOld to current values
        for comp in range(6):
            rOld[obj][comp] = r[obj][comp]

    for event in pygame.event.get():
        if event.type == QUIT:
            plotting = False
        if event.type == KEYDOWN and event.key == K_ESCAPE:
            plotting = False

# Draw circles at final points
for obj in range(n):
    pygame.draw.circle(screen,lineColor[obj],   \
        (math.floor(xCoord(r[obj][0])),math.floor(yCoord(r[obj][1]))),3)
pygame.display.flip()

endTime = datetime.now()

EFinal = KE()+PE()      # Compute final energy

print("Time elapsed:",endTime-startTime)
print("tFinal   =",t)
print("EInitial =",EInitial)
print("EFinal   =",EFinal)
print("EError   =",(EFinal-EInitial)/EInitial*100)

running = True
while running == True:
    for event in pygame.event.get():
        if event.type == QUIT:
            running = False
        if event.type == KEYDOWN and event.key == K_ESCAPE:
            running = False

pygame.display.quit()


