# n-body electrostatic force simulator
# Simulates a system of n charges in space.
# The mass of each object is the absolute value of the charge
# by William Ho

import math, pygame, random
from pygame.locals import *
from datetime import datetime

#######################
# Values of constants #
#######################

defaultFileName = "efield_input.txt"

# Define variables for graphical output
width = height = 500
bgColor = (255,255,255)
axisColor = (128,128,128)
scale = 5

########################
# Function definitions #
########################

def fileLength(fileName):
    """Returns the number of lines in a file"""
    file = open(fileName,"r")
    
    totalLines = 0
    for line in file:
        totalLines += 1
    file.close()
    
    return totalLines+1

def dInput(string,default):
    """A version of the input() function with a default value if
        nothing is entered"""
    print(string," [",default,"]:",sep="",end=" ")
    inputValue = input()
    if inputValue == "":
        inputValue = default
    else:
        inputValue = inputValue
    return inputValue
    
def calcForce(obj1,obj2):
    """Returns the abs.value of the force between two objects.
       In terms of an equation:   F = (q1*q2) * abs(r_x1 - r_x2) / r^3"""
    # Calculate the inverse of the distance between two points, cubed
    rCubed = ((rTemp[obj2][0]-rTemp[obj1][0])**2        \
              +(rTemp[obj2][1]-rTemp[obj1][1])**2       \
              +(rTemp[obj2][2]-rTemp[obj1][2])**2       \
              )**(-1.5)

    # Calculate the abs value of the numerator (r_x1 - r_x2) with offset
    numerator = [0.0]*3 
    for comp in range(3):   # Set values for x, y, z components of numerator
        numerator[comp] = (q[obj1]*q[obj2])*abs(rTemp[obj1][comp]-rTemp[obj2][comp])
    return numerator[0]*rCubed,numerator[1]*rCubed,numerator[2]*rCubed

def forceArray():
    """Stores all the unique forces in an array"""
    for obj1 in range(n):
        for obj2 in range(n):
            if obj1 < obj2:
                # Since the force of 1 on 2 is equal in magnitude
                # to the force of 2 on 1, only one of these is stored
                force[0][obj1][obj2],    \
                force[1][obj1][obj2],    \
                force[2][obj1][obj2] =   \
                calcForce(obj1,obj2)

def acc(obj1, comp):
    """Returns the component of acceleration for the specified object"""
    # Calculate the total force divided by the mass of the specified object
    netFOverM = 0
    for obj2 in range(n):
        # Determine the sign of the force based on relative positions
        if rTemp[obj1][comp] > rTemp[obj2][comp]:
            posOrNeg = 1
        elif rTemp[obj1][comp] < rTemp[obj2][comp]:
            posOrNeg = -1
        elif rTemp[obj1][comp] == rTemp[obj2][comp]:
            posOrNeg = 0

        # Add to the total F/m
        if obj1 > obj2:
            netFOverM += posOrNeg * force[comp][obj2][obj1] / m[obj1]
        elif obj1 < obj2:
            netFOverM += posOrNeg * force[comp][obj1][obj2] / m[obj1]
    return netFOverM

def RK4():
    """Updates values for position and velocity using Runge-Kutta 4"""
    for slopeNum in range(4):
        # offset represents the value that is multiplied by the RKSlope
        if slopeNum == 0:
            offset = 0
        elif slopeNum == 1 or slopeNum == 2:
            offset = 0.5*h
        elif slopeNum == 3:
            offset = h

        for obj in range(n):
            for comp in range(6):
                if movable[obj] == 1:
                    rTemp[obj][comp] = r[obj][comp]+offset*RKSlope[slopeNum-1][obj][comp]
                else:
                    rTemp[obj][comp] = r[obj][comp]
        
        # Determine the RKSlope for each object
        for obj in range(n):
            forceArray()
            for comp in range(3):
                RKSlope[slopeNum][obj][comp] = movable[obj]*r[obj][comp+3]+offset*RKSlope[slopeNum-1][obj][comp+3]
                RKSlope[slopeNum][obj][comp+3] = movable[obj]*acc(obj,comp)

    # Update all values for position and velocity
    for obj in range (n):
        if movable[obj] != 0:
            for comp in range(6):
                r[obj][comp] += hConst*(RKSlope[0][obj][comp] \
                                +2*RKSlope[1][obj][comp] \
                                +2*RKSlope[2][obj][comp] \
                                +RKSlope[3][obj][comp])

def xCoord(xValue):
    """Converts the value of x into an x value on the graph"""
    return scale*xValue+xCenter

def yCoord(yValue):
    """Converts the value of y into a y value on the graph"""
    return -scale*yValue+yCenter

def KE():
    """Returns the total kinetic energy in the system at the current time"""
    totalKE = 0
    for obj in range(n):
        totalKE += .5*m[obj]*(r[obj][3]**2+r[obj][4]**2+r[obj][5]**2)
    return totalKE

def PE():
    """Returns the total potential energy in the system at the current time"""
    totalPE = 0
    for obj1 in range(n):
        for obj2 in range(n):
            if obj1 < obj2:
                distInv = 1/((r[obj2][0]-r[obj1][0])**2     \
                          +(r[obj2][1]-r[obj1][1])**2       \
                          +(r[obj2][2]-r[obj1][2])**2)**.5
                totalPE -= (m[obj1]*m[obj2])*distInv
    return totalPE

def elecPotential():
    """Returns the total electric potential energy in the system at the current time"""
    totalEPotential = 0
    for obj1 in range(n):
        for obj2 in range(n):
            if obj1 < obj2:
                distInv = 1/((r[obj2][0]-r[obj1][0])**2     \
                          +(r[obj2][1]-r[obj1][1])**2       \
                          +(r[obj2][2]-r[obj1][2])**2)**.5
                totalEPotential += (q[obj1]*q[obj2])*distInv
    return totalEPotential

########
# Main #
########

fileName = dInput("Name of file",defaultFileName)
print("(Plotting of the solution curve can be stopped by pressing the ESC key)")
file = open(fileName,"r")

h = float(file.readline().split("#")[0])        # Step size
hConst = h/6                                    # Constant used in RK4
tFinal = float(file.readline().split("#")[0])   # Final time

n = math.floor((fileLength(fileName)-2)/4)      # Number of objects
m = [0.0]*n
q = [0.0]*n
movable = [0]*n
lineColor = [0]*n

for obj in range(n):
    lineColor[obj] = (random.randint(0,255),random.randint(0,255),random.randint(0,255))

# r array consists of [obj 1, obj 2, ... , obj n][r_x, r_y, r_z, v_x, v_y, v_z]
rOld = [[0.0 for component in range(6)] for objNumber in range(n)]
r = [[0.0 for component in range(6)] for objNumber in range(n)]
rTemp = [[0.0 for component in range(6)] for objNumber in range(n)]

# Read values from the input file and insert them into the appropriate arrays
for i in range(n):
    q[i] = float(file.readline().split("#")[0])     # Read values for charges
    m[i] = abs(q[i])                                # Mass is abs. value of charge
    movable[i] = int(file.readline().split("#")[0]) # Is this object movable?
    for j in range(2):
        rOrV = j*3      # if 0, this will refer to position; if 1, velocity
        rOld[i][rOrV],rOld[i][rOrV+1],rOld[i][rOrV+2] = file.readline().split("#")[0].split(", ")     # Read values for r_x, r_y, r_z, v_x, v_y, v_z
        for component in range(3):
            rOld[i][rOrV+component] = float(rOld[i][rOrV+component])

file.close()

# Set r array equal to rOld array
for obj in range(n):
    for comp in range(6):
        r[obj][comp] = rTemp[obj][comp] = rOld[obj][comp]

# Initialize global arrays that will be used in functions
force = [[[0.0 for obj2 in range(n)] for obj1 in range(n)]        \
           for component in range(3)]

RKSlope = [[[0.0 for comp in range(6)] for objNum in range(n)]      \
           for slopeNum in range(4)]

# Load a screen for graphical output
pygame.init()
screen = pygame.display.set_mode((width,height))
screen.fill(bgColor)
xCenter = width/2
yCenter = height/2
pygame.draw.line(screen,axisColor,(yCenter,0),(yCenter,width),1)
pygame.draw.line(screen,axisColor,(0,xCenter),(height,xCenter),1)

EInitial = KE()+elecPotential()   # Compute initial energy

startTime = datetime.now()

# Draw circles at initial points
for obj in range(n):
    if movable[obj] == 0:
        pygame.draw.circle(screen,lineColor[obj],   \
            (math.floor(xCoord(r[obj][0])),math.floor(yCoord(r[obj][1]))),3)

# Plot solution curves to the screen
t = 0
plotting = True
while t < tFinal + hConst and plotting == True:
    RK4()   # Update position and velocity for all objects
    for obj in range(n):
        pygame.draw.line(screen,lineColor[obj],
            (xCoord(rOld[obj][0]), yCoord(rOld[obj][1])), \
            (xCoord(r[obj][0]),yCoord(r[obj][1])),1)
    pygame.display.flip()
    t += h

    for obj in range(n):    # Update values in rOld to current values
        for comp in range(6):
            rOld[obj][comp] = r[obj][comp]

    for event in pygame.event.get():
        if event.type == QUIT:
            plotting = False
        if event.type == KEYDOWN and event.key == K_ESCAPE:
            plotting = False

# Draw circles at final points
for obj in range(n):
    if movable[obj] != 0:
        pygame.draw.circle(screen,lineColor[obj],   \
            (math.floor(xCoord(r[obj][0])),math.floor(yCoord(r[obj][1]))),3)
pygame.display.flip()

endTime = datetime.now()

EFinal = KE()+elecPotential()      # Compute final energy

print("Time elapsed:",endTime-startTime)
print("tFinal   =",t)
print("EInitial =",EInitial)
print("EFinal   =",EFinal)
print("EError   =",(EFinal-EInitial)/EInitial*100)

running = True
while running == True:
    for event in pygame.event.get():
        if event.type == QUIT:
            running = False
        if event.type == KEYDOWN and event.key == K_ESCAPE:
            running = False

pygame.display.quit()
