# Bead on a wire
# Simulates the motion of a bead placed on a wire

import pygame, math, os, random
from pygame.locals import *
from math import *

#######################
# Values of constants #
#######################

# Default values if text file not found
xEquation = "t"
yEquation = "3*math.sin(t)"
tArrayInitial = -2*pi
tArrayFinal = 2*pi
arrayEntries = 5000

# Initial conditions for the bead
pos = 0         # Initial position of the bead (arc length)
vel = 0         # Initial velocity of the bead
frict = 0       # Friction
h = .01          # Step size

# Constants used for graphical output
width = 550
height = 400
padding = 0.15   # Empty space at the edges of the screen.
                #   This value should be much less than 0.5

bgColor = (255,255,255)
lineColor = (0,0,0)

beadColor = (255,0,0)
beadRadius = 5      # Radius of the bead, in pixels

fontName = "Courier New"
fontSize = 15
fontColor = (0,0,0)

gridColumns = 3
gridRows = height/fontSize

showInfo = False    # Data regarding pos/vel/etc are hidden initially

# Load curves from a file
fileName = "bead_input.txt"

########################
# Function definitions #
########################

def fileLength(fname):
    """Returns the number of lines in a file. From:
        http://stackoverflow.com/questions/845058/how-to-get- 
        line-count-cheaply-in-python"""
    with open(fname) as f:
        for i, l in enumerate(f):
            pass
    return i + 1

def xCalc(t):
    """Parametric equation #1"""
    return eval(xEquation)

def yCalc(t):
    """Parametric equation #2"""
    return eval(yEquation)

def derivative(comp):
    """Estimate derivatives of the pair of parametric equations
        using the five-point stencil method
        (http://en.wikipedia.org/wiki/Five-point_stencil)"""
    if comp == "x":
        return (-xCalc(t+2*tArrayInc)+8*xCalc(t+tArrayInc)-8* \
              xCalc(t-tArrayInc)+xCalc(t-2*tArrayInc))*(1/12*1/tArrayInc)
    if comp == "y":
        return (-yCalc(t+2*tArrayInc)+8*yCalc(t+tArrayInc)-8* \
              yCalc(t-tArrayInc)+yCalc(t-2*tArrayInc))*(1/12*1/tArrayInc)
    
def getIndex(t):
    """For purposes of plotting the curve, inputting the parameter t returns
        the corresponding position in the x, y, posArray, and k lists."""
    return math.floor((t-tArrayInitial)*tArrayEntries)

def getTime(index):
    """For purposes of plotting the curve, inputting the position of an entry
        in the x, y, posArray, or k lists outputs the associated t."""
    return tArrayInitial + index*tArrayInc

def dist(x1,y1,x2,y2):
    """Returns distance between two points defined by x and y values."""
    return math.sqrt((x2-x1)**2+(y2-y1)**2)

def binarySearch(value,array,low=None,high=None):
    """Binary search tree used to find the index of an entry in a list
        (Only works effectively with an ordered list)"""
    if low == None:
        low = 0
    if high == None:
        high = len(array)

    while low < high:
        mid = (low+high)//2
        midValue = array[mid]
        if midValue < value:
            low = mid+1
        elif midValue > value: 
            high = mid
        else:
            return mid
    return mid

def unorderedSearch(value,array,condition=None):
    """Search through an unordered list and return a list of indices in the
        list that are closest to the specified value"""
    indexMatches = list()
    for index in range(len(array)-1):
        currentValue = array[index]
        nextValue = array[index+1]
        if condition == None:
            if currentValue > value > nextValue  or     \
               currentValue < value < nextValue:
                    indexMatches.append(index)
        if condition == "below":
            if value > currentValue:
                    indexMatches.append(index)
        if condition == "above":
            if value < currentValue:
                    indexMatches.append(index)
    return indexMatches

def matchesInList(list1,list2):
    """Returns the indices of elements that are present in both lists"""
    return list(set(list1).intersection(list2))

def findClosest(xValue,yValue,currentIndex=0,error=0.001):
    """Search through an unordered list and return the index in the
        list if a position within a certain range of the specified value"""
    indexMatches = list()
    for index in range(arrayEntries):
        if xValue + error > x[index] > xValue - error  and     \
           yValue + error > y[index] > yValue - error:
            if index != currentIndex:
                indexMatches.append(index)

    if len(indexMatches) != 0:
        minDist = 1e10
        closestIndex = None
        currentDist = 0
        for index in indexMatches:
            if index+arrayEntries/100 > currentIndex > index-arrayEntries/100:
                continue
            else:
                currentDist = dist(xValue,yValue,x[index],y[index])
                if currentDist < minDist:
                    minDist = currentDist
                    closestIndex = index
        return closestIndex

def xyMax(XOrY,indexList):
    """Returns the max x or y value in a list of indices
        (To be used with the matchesInList function)"""
    maxValue = max([XOrY[index] for index in indexMatches])
    for index in indexList:
        if XOrY[index] == maxValue:
            return index

def xyMin(XOrY,indexList):
    """Returns the min x or y value in a list of indices
        (To be used with the matchesInList function)"""
    minValue = min([XOrY[index] for index in indexMatches])
    for index in indexList:
        if XOrY[index] == minValue:
            return index

def returnToCurve():
    """Checks to see if the bead returns to the curve
        (If it is currently doing projectile motion)"""
    xMatches = list()
    yMatches = list()
    for index in range(len(x)):
        xValue = x[index]
        if beadX > xValue > beadXOld or beadX < xValue < beadXOld:
            xMatches.append(index)
        yValue = y[index]
        if beadY > yValue > beadYOld or beadY < yValue < beadYOld:
            yMatches.append(index)        
    indexMatches = matchesInList(xMatches,yMatches)
    if len(indexMatches) != 0:
        minDist = 1e10
        currentDist = 0
        closestIndex = 0
        for index in indexMatches:
            currentDist = dist(beadXOld,beadYOld,x[index],y[index])
            if currentDist < minDist:
                minDist = currentDist
                closestIndex = index
        return index
        
def acc(pos,vel):
    """Returns the acceleration of the bead (if on the curve)"""
    index = binarySearch(pos,posArray)
    return -k[index]/math.sqrt(1+k[index]**2)-frict*vel     

def RK4(pos,vel):
    """Updates position and velocity of the bead (if on the curve)"""
    pSlope1 = vel
    vSlope1 = acc(pos,vel)
    pSlope2 = vel + .5*h*vSlope1
    vSlope2 = acc(pos+.5*h*pSlope1,pSlope2)
    pSlope3 = vel + .5*h*vSlope2
    vSlope3 = acc(pos+.5*h*pSlope2,pSlope3)
    pSlope4 = vel + h*vSlope3
    vSlope4 = acc(pos+h*pSlope3,pSlope4)

    pos += hConst*(pSlope1 + 2*pSlope2 + 2*pSlope3 + pSlope4)
    vel += hConst*(vSlope1 + 2*vSlope2 + 2*vSlope3 + vSlope4)

    return pos,vel

def xCoord(xValue):
    """Converts the value of x into an x value on the graph"""
    return scale*xValue+xOrigin

def yCoord(yValue):
    """Converts the value of y into a y value on the graph"""
    return -scale*yValue+yOrigin

def xTrue(xOnScreen):
    """Returns the 'true' x coordinate of a point on the screen
        based on the parameters of the parametric curve"""
    return (xOnScreen - xOrigin)/scale

def yTrue(yOnScreen):
    """Returns the 'true' y coordinate of a point on the screen
        based on the parameters of the parametric curve"""
    return -(yOnScreen-yOrigin)/scale

def KE():
    """Returns the kinetic energy of the bead"""
    if onCurve == True:
        return 0.5*vel**2
    else:
        return 0.5*(v_x**2+v_y**2)

def PE():
    """Returns the height of the bead above the origin"""
    if onCurve == True:
        index = binarySearch(pos,posArray)
        return y[index]
    else:
        return beadY

def displayText(text,coord,color=fontColor,size=fontSize,surface=None):
    """Blit text to the specified surface"""
    font = pygame.font.Font(pygame.font.match_font(fontName), size)
    showText = font.render(text, 1, color)
    textPos = coord
    if surface == None:
        screen.blit(showText, textPos)
    else:
        surface.blit(showText,textPos)

def centeredText(text,yTop,color=fontColor,size=fontSize,surface=None):
    font = pygame.font.Font(pygame.font.match_font(fontName), size)
    showText = font.render(text, 1, color)
    textPos = showText.get_rect(top=yTop,centerx=width/2)
    if surface == None:
        screen.blit(showText,textPos)
    else:
        surface.blit(showText,textPos)

def drawEnergyLine(surface):
    """Draw a horizontal line representing initial total energy"""
    pygame.draw.aaline(surface,beadColor,(0,yCoord(EInitial)), \
                       (width,yCoord(EInitial)),1)
    energyText = font.render("EInitial", 1, beadColor)
    energyTextRect = energyText.get_rect(bottomright=(width,yCoord(EInitial)))
    surface.blit(energyText, energyTextRect)

    energyText2 = font.render("% .3f" % (EInitial), 1, beadColor)
    energyText2Rect = energyText.get_rect(topright=(width,yCoord(EInitial)))
    surface.blit(energyText2, energyText2Rect)

def clearBead():
    """Blit a copy of the curve surface to the screen"""
    screen.blit(curve,(0,0))

def updateBead():
    """Blit a copy of the bead surface to the appropriate location on screen"""
    if onCurve == True:
        index = binarySearch(pos,posArray)
        screen.blit(bead,(math.floor(xCoord(x[index]))-beadRadius, \
                          math.floor(yCoord(y[index]))-beadRadius))
    else:
        screen.blit(bead,(math.floor(xCoord(beadX)-beadRadius), \
                          math.floor(yCoord(beadY)-beadRadius)))

def grid(column,row):
    """Converts positions in a grid to coordinates on the screen"""
    return column*gridWidth,row*gridHeight

def showData():
    """Display info about the bead (position, energy, etc) on the screen"""
    if showInfo == True:
        centeredText("Press SPACE to hide data, ESC to restart",grid(0,0)[1])
        displayText("ETotal:% .3f" % (KE()+PE()),grid(2,1))
        displayText("EError:% .3f%%" % (EError),grid(2,2))
        displayText("PE:% .3f" % (PE()),grid(1,1))
        displayText("KE:% .3f" % (KE()),grid(1,2))

        if onCurve == True:
            displayText("pos:% .3f" % (pos),grid(0,1))
            displayText("vel:% .3f" % (vel),grid(0,2))
        else:
            displayText("pos(x):% .3f" % (beadX),grid(0,1))            
            displayText("pos(y):% .3f" % (beadY),grid(0,2))
            displayText("vel(x):% .3f" % (v_x),grid(0,3))            
            displayText("vel(y):% .3f" % (v_y),grid(0,4))
            displayText("vel   :% .3f" % (math.sqrt(v_x**2+v_y**2)),grid(0,5))
            
        displayText("time: %.1f" % (t),(1,height-fontSize))
    else:
        centeredText("Press SPACE to show data, ESC to restart",grid(0,0)[1])


########
# Main #
########

pygame.init()
font = pygame.font.Font(pygame.font.match_font(fontName), fontSize)

# Partition the screen (for text display)
gridWidth = round(width/gridColumns)
gridHeight = round(height/gridRows)

# Load a screen for graphical output
screen = pygame.display.set_mode((width,height))
pygame.display.set_caption("Bead on a Wire")
screen.fill(bgColor)

offset = 0
while True:
    choosingCurve = True
    selecting = True
    if os.path.exists(fileName):
        file = open(fileName,"r")
        numCurves = round((fileLength(fileName))/5)    # Number of pairs of equations
        eqArray = [[0.0 for fields in range(5)] for curves in range(numCurves)]

        h = float(file.readline().split("#")[0])
        arrayEntries = int(file.readline().split("#")[0])
        file.readline()

        for i in range(numCurves):
            for j in range(3):
                eqArray[i][j] = file.readline().split("\n")[0].split("#")[0].split("\t")[0]
            eqArray[i][3],eqArray[i][4] = file.readline().split("\n")[0]. \
                                          split("#")[0].split("\t")[0].split(",")
            file.readline()
        
        while choosingCurve == True:
            screen.fill(bgColor)
            centeredText("Select a curve from the list using the number keys.",grid(0,0)[1])
            centeredText("Use the LEFT and RIGHT keys to switch between pages.",grid(0,1)[1])

            centeredText("< Page %d of %d >" % (offset+1,math.ceil(numCurves/5)), \
                        grid(0,gridRows-1)[1])
            
            for i in range(offset*5,offset*5+5):
                if i < numCurves:
                    eqIndex = i%5
                    displayText(" %d. %s (%s < t < %s)" % (eqIndex+1,eqArray[i][0],    \
                                        eqArray[i][3],eqArray[i][4]),grid(0,eqIndex*4+4))
                    displayText("    x = %s" % (eqArray[i][1]),grid(0,eqIndex*4+5))
                    displayText("    y = %s" % (eqArray[i][2]),grid(0,eqIndex*4+6))
            pygame.display.flip()
            
            for event in pygame.event.get():
                if event.type == QUIT:
                    os._exit(99)
                if event.type == KEYDOWN:
                    if event.key == K_ESCAPE:
                        os._exit(99)
                    if event.key == K_LEFT:
                        if offset > 0:
                            offset -= 1
                        else:
                            offset = math.ceil(numCurves/5)-1
                    if event.key == K_RIGHT:
                        if offset*5+5 < numCurves:
                            offset += 1
                        else:
                            offset = 0
                    for numKey in range(49,54):
                        if event.key == numKey and numKey-49 < numCurves:
                            chosenCurve = (numKey - 49) + offset*5
                            choosingCurve = False

        xEquation = eqArray[chosenCurve][1]
        yEquation = eqArray[chosenCurve][2]
        tArrayInitial = eval(eqArray[chosenCurve][3])
        tArrayFinal = eval(eqArray[chosenCurve][4])

    hConst = h/6

    tArrayInc = (tArrayFinal - tArrayInitial)/arrayEntries

    x = [0.0]*arrayEntries
    y = [0.0]*arrayEntries
    posArray = [0.0]*arrayEntries
    k = [0.0]*arrayEntries

    screen.fill(bgColor)
    plottingText = font.render("Plotting curve...", 1, fontColor)
    plottingTextRect = plottingText.get_rect(centerx=width/2,centery=height/2)
    screen.blit(plottingText, plottingTextRect)
    pygame.display.flip()

    # Initialize coordinate, position, and curve slope arrays
    for now in range(arrayEntries):
        prev = now-1
        t = getTime(now)
        x[now] = xCalc(t)
        y[now] = yCalc(t)

        xSlope = derivative("x")
        ySlope = derivative("y")

        if xSlope == 0:             # To avoid division by zero errors
            xSlope,ySlope = 1,1     #   vertical slopes will have a slope of 1

        if now == 0:     # Check if this point is the first point on the curve
            posArray[now] = 0
            if x[now+1] == x[now]:
                k[now] = -1
            else:
                k[now] = ySlope/xSlope
                
        else:   # Set values for position and slope arrays
            posArray[now] = posArray[prev] + dist(x[prev],y[prev],x[now],y[now])
            if x[now] > x[prev]:
                k[now] = ySlope/xSlope
            elif x[now] < x[prev]:          # If the curve is heading to the left,
                k[now] = -ySlope/xSlope     #   reverse the sign of the slope.
            else:
                k[now] = -1     # To avoid a division by zero error
            
    startEquiv = findClosest(x[0],y[0],0)
    endEquiv = findClosest(x[arrayEntries-1],y[arrayEntries-1],arrayEntries-1)

    # Autoscaling
    xMin = min(x)
    xMax = max(x)
    yMin = min(y)
    yMax = max(y)

    maxValue = max(abs(xMax)+abs(xMin),abs(yMax)+abs(yMin))
    minAxis = min(width,height-fontSize*2)
    scale = minAxis/maxValue*(1-padding)
    xOrigin = width/2 - scale*(xMin+xMax)/2
    yOrigin = height/2 + scale*(yMin+yMax)/2 + fontSize

    # Create a surface for the bead
    bead = pygame.Surface((beadRadius*2,beadRadius*2))
    bead.fill(bgColor)
    pygame.draw.circle(bead,beadColor,(beadRadius,beadRadius),beadRadius)

    # Create a surface for the curve
    curve = pygame.Surface((width,height))
    curve.fill(bgColor)

    # Draw the curve
    for now in range(arrayEntries):
        prev = now-1
        t = getTime(now)
        if now != 0:
            pygame.draw.aaline(curve,lineColor,(xCoord(x[prev]),yCoord(y[prev])), \
                     (xCoord(x[now]),yCoord(y[now])),1)

    screen.blit(curve,(0,0))    
    pygame.display.flip()

    # Selecting where to place the bead
    afterChoosingCurve = True
    while afterChoosingCurve == True:
        pos = 0
        vel = 0
        onCurve = True
        paused = False
        updating = True
        while selecting == True:
            # Get mouse x,y and convert it to "true" x,y in terms of equations
            mousePos = pygame.mouse.get_pos()
            xMouse = xTrue(mousePos[0])
            yMouse = yTrue(mousePos[1])
            
            clearBead()
            centeredText("Select a starting position for the bead.",grid(0,0)[1])
            centeredText("Press UP or DOWN to change initial velocity.",grid(0,1)[1])
            centeredText("Press LEFT or RIGHT to change friction.",grid(0,2)[1])
            centeredText("Press ESC to return to previous screen.",grid(0,3)[1])
            displayText("pos: %.3f" % (pos),grid(0,gridRows-1))
            displayText("vel: %.1f" % (vel),grid(1,gridRows-1))
            displayText("frict: %.3f" % (frict),grid(2,gridRows-1))

            # Define left and right bounds for selection. If too far to
            #   the left, it uses the leftmost point. If too far to the
            #   right, it uses the rightmost point.
            if xMouse < xMin:
                xMouse = xMin+1e-10
            elif xMouse > xMax:
                xMouse = xMax-1e-10

            # Search through the arrays to find the closest values for x and y
            xIndices = unorderedSearch(xMouse,x)
            yIndices = unorderedSearch(yMouse,y,"below")
            indexMatches = matchesInList(xIndices,yIndices)

            # If there are no matching y values below the cursor, find the lowest
            #   y value on the curve above the cursor
            if len(indexMatches) == 0:
                yIndices = unorderedSearch(yMouse,y,"above")
                indexMatches = matchesInList(xIndices,yIndices)
                if len(indexMatches) != 0:
                    index = xyMin(y,indexMatches)
            else:
                index = xyMax(y,indexMatches)

            pos = posArray[index]   # Set position based on x and y coordinates
                
            EInitial = KE()+PE()
            drawEnergyLine(screen) # Draw a horizontal line representing total energy
            
            updateBead()
            pygame.display.flip()

            # Checks which keys are held down
            key = pygame.key.get_pressed()
            if key[K_UP]:
                vel += .1
            if key[K_DOWN]:
                vel -= .1
            if key[K_LEFT]:
                frict -= .001
            if key[K_RIGHT]:
                frict += .001

            if frict < 0:
                frict = 0

            # Check which keys are pressed
            for event in pygame.event.get():
                if event.type == MOUSEBUTTONDOWN:
                    selecting = False
                if event.type == QUIT:
                    os._exit(99)
                if event.type == KEYDOWN:
                    if event.key == K_ESCAPE:
                        selecting = False
                        updating = False
                        afterChoosingCurve = False
                    if event.key == K_RETURN:
                        selecting = False

        EInitial = KE()+PE()
        if EInitial == 0:       # To avoid division by zero in error calculations
            EInitial = 1e-10

        pauseText = font.render("Press P to pause", 1, fontColor)
        pauseTextRect = pauseText.get_rect(bottomright=(width,height))

        EError = 0
        # Update the position and velocity of the bead with RK4
        while updating == True:
            clearBead()
            drawEnergyLine(screen) # Draw a horizontal line representing total energy
            screen.blit(pauseText, pauseTextRect)

            ECurrent = KE()+PE()
            EError = (ECurrent-EInitial)/EInitial*100
            
            if pos != None:
                currentIndex = binarySearch(pos,posArray)
 
                # Check to see if the bead has left the curve
                if pos <= 0 or pos >= posArray[arrayEntries-1] and vel > 1e-5:
                    if pos <= 0:
                        if startEquiv != None:
                            pos = posArray[startEquiv]
                        else:
                            prevIndex = currentIndex+1
                            onCurve = False
                    elif pos >= posArray[arrayEntries-1]:
                        if endEquiv != None:
                            pos = posArray[endEquiv]
                        else:
                            prevIndex = currentIndex-1
                            onCurve = False
                 
                if onCurve == False:
                    # Is the projectile moving to the left or right?
                    if x[currentIndex] > x[prevIndex]:
                        xDirection = 1
                    elif x[currentIndex] < x[prevIndex]:
                        xDirection = -1
                    else:
                        xDirection = 0

                    # Is the projectile moving up or down?
                    if y[currentIndex] > y[prevIndex]:
                        yDirection = 1
                    elif y[currentIndex] < y[prevIndex]:
                        yDirection = -1
                    else:
                        yDirection = 0

                    pos = None          # A pos value of 'None' = off the curve
                    beadX = x[currentIndex]     # Set current x location of bead
                    beadXOld = beadX
                    beadY = y[currentIndex]     # Set current y location of bead
                    beadYOld = beadY

                    # Find the x and y components of velocity based on the angle
                    #   of the tangent line to the curve at that point.
                    sinOfAngle = abs(k[currentIndex]/math.sqrt(1+k[currentIndex]**2))
                    v_x = xDirection*abs(vel)*math.cos(math.asin(sinOfAngle))
                    v_y = yDirection*abs(vel)*sinOfAngle    
            
            if onCurve == True:
                pos,vel = RK4(pos,vel)  # If bead is on the curve, resume RK4

            else:                       # If not on the curve, projectile motion.
                beadX += v_x*h 
                beadY += v_y*h
                index = returnToCurve()     # Check if the bead returns to the curve
                
                if index != None:
                    onCurve = True
                    pos = posArray[index]

                    if v_x > 0:
                        if x[index] > x[index-1]:
                            xSign = 1
                        if x[index] < x[index-1]:
                            xSign = -1

                    elif v_x < 0:
                        if x[index] > x[index-1]:
                            xSign = -1
                        if x[index] < x[index-1]:
                            xSign = 1

                    vel = xSign*math.sqrt(v_x**2+v_y**2)

                    # Because the landing position is not exact, adjustments
                    #   to velocity must be made to conserve energy
                    while EError > 0.1:
                        if vel < 0:
                            vel += 0.01
                        else:
                            vel -= 0.01
                        EError = (KE()+PE()-EInitial)/EInitial*100
                    while EError < -0.1:
                        if vel > 0:
                            vel += 0.01
                        else:
                            vel -= 0.01
                        EError = (KE()+PE()-EInitial)/EInitial*100
                    
                v_y -= h               # Update vertical velocity of the bead
                beadXOld,beadYOld = beadX,beadY
            
            updateBead()
            showData()

            pygame.display.flip()
            t += h
            
            for event in pygame.event.get():
                if event.type == QUIT:
                    os._exit(99)
                if event.type == KEYDOWN:
                    if event.key == K_SPACE:
                        showInfo = not showInfo
                    if event.key == K_ESCAPE:
                        updating = False
                        selecting = True
                    if event.key == K_p:
                        # If P is pressed, the program is paused
                        paused = True

            while paused:
                # Halt updating and display the word "Paused" on the screen
                clearBead()
                pausedText = font.render("Paused", 1, fontColor)
                pausedTextRect = pausedText.get_rect(centerx=width/2,centery=height/2)
                screen.blit(pausedText, pausedTextRect)
                showData()
                updateBead()
                drawEnergyLine(screen)
                pygame.display.flip()
                for event in pygame.event.get():            
                    if event.type == QUIT:
                        os._exit(99)
                    if event.type == KEYDOWN:
                        if event.key == K_SPACE:
                            showInfo = not showInfo
                        if event.key == K_ESCAPE:
                            updating = False
                            selecting = True
                            paused = False
                        if event.key == K_p:
                            paused = False
