import pygame, math, os, random
from pygame.locals import *
from datetime import datetime
from math import *

#######################
# Values of Constants #
#######################

# Screen constants
SCREEN_WIDTH =  768
SCREEN_HEIGHT = 608
SCREEN_CENTERX =  SCREEN_WIDTH/2

# Size constants
SQUARE_SIZE =   64          # Length/width of one tile on the battlefield
UNIT_SIZE =     32          # Length/width of a unit on the battlefield

# Text constants
FONT_NAME =     "Trebuchet MS"
TEXT_PADDING =  12 
FONT_SIZE =     20
FONT_SIZE2 =    32

# Title screen constants
TITLE = "Diff EQ Wars"
titleSections = ["Randomize","Load Map","Edit Map","Help","Quit"]
titleInfo = ["Generates a random scenario to simulate",
             "Load a previously-created map",
             "Create your own custom map",
             "Information about how the game works",
             "Close the game"]
TITLE_BOX_POS = (50,365)
TITLE_TEXT_POS = (TITLE_BOX_POS[0]+UNIT_SIZE+TEXT_PADDING,TITLE_BOX_POS[1]+TEXT_PADDING)
TITLE_BG_WIDTH = SCREEN_WIDTH*3
TITLE_UNIT_NUM = 30
TITLE_FOREST_NUM = 100
TITLE_MOUNTAIN_NUM = 40
TITLE_BASE_NUM = 4
TITLE_SCROLL_SPEED = .3

# Locations of files
SETTINGS_FILE =     "settings.txt"
MAPS_FILE =         "maps.txt"
HELP_MENU_FILE =    "help/help_text.txt"

# Help screen constants
HELP_CONTENTS_POS = (45,115)
HELP_MAINTEXT_POS = (230,20)

# Color constants
BG_COLOR =      0xFFFFFF
TRANS_COLOR =   0xFF00FF

FRIENDLY_COLOR =    0x0000FF
FRIENDLY_COLOR2 =   0x0099FF
HOSTILE_COLOR =     0xFF0000
SUPPLY_COLOR =      0x00FF00

TEXT_COLOR = (0,0,0)
TEXT_COLOR_INV = (255-TEXT_COLOR[0],255-TEXT_COLOR[1],255-TEXT_COLOR[2])

# Battle simulation constants
STEP_SIZE = .1
HCONST =    STEP_SIZE/6

ROWS = 7        # Battlefield dimensions (by number of panels)
COLUMNS = 12
BOUNDARY_RIGHT = SQUARE_SIZE*COLUMNS
BOUNDARY_BOTTOM = SQUARE_SIZE*ROWS
TEXTBOX_CENTERY = SCREEN_HEIGHT/2+BOUNDARY_BOTTOM/2-UNIT_SIZE/2

WP_RADIUS = 7               # Radius of the waypoint circle graphic
WP_COLOR = FRIENDLY_COLOR   # Waypoint color

BASE_REINFORCEMENT = 50     # Reinforcement rate for units at a allied base
DIST_FROM_BASE = 100        # Starting distance from base for randomized units

# Speed multipliers for various tiles
SPEED_NORMAL = 1
SPEED_FOREST = .5
SPEED_MOUNTAIN = .3

########################
# Function Definitions #
########################

#########################
# Title Screen Functions
def titleContents(strings):
    """Display the table of contents of the main menu"""
    verticalOffset = TITLE_TEXT_POS[1]
    for line in strings:
        text = font2.render(line, 1, TEXT_COLOR)
        screen.blit(text,(TITLE_TEXT_POS[0],verticalOffset))
        verticalOffset += FONT_SIZE2

def showCursor(pos,verticalOffset):
    """Display a cursor showing where the user is in the table of contents"""
    screen.blit(cursorSurface, (pos[0]-UNIT_SIZE,pos[1]+verticalOffset*UNIT_SIZE+5), (0,0,32,32))

def centeredText(string,y,large=False,outline=0):
    """Prints horizontally-centered text on the screen, with an outline"""
    if large == True:
        whichFont = font2
    else:
        whichFont = font
    text = whichFont.render(string, 1, TEXT_COLOR_INV)
    textRect = text.get_rect(centery=y,centerx=SCREEN_CENTERX)
    if outline != 0:
        screen.blit(text,(textRect[0]-outline,textRect[1]))
        screen.blit(text,(textRect[0]+outline,textRect[1]))
        screen.blit(text,(textRect[0],textRect[1]-outline))
        screen.blit(text,(textRect[0],textRect[1]+outline))
    text = whichFont.render(string, 1, TEXT_COLOR)
    screen.blit(text,textRect)

def titleBlit():
    """Blit the title background to the screen"""
    screen.blit(titleBGSurface,(titleOffset,0))
    if titleOffset < -TITLE_BG_WIDTH+SCREEN_WIDTH:
        screen.blit(titleBGSurface,(titleOffset+TITLE_BG_WIDTH,0))

def titleScreenUpdate():
    """Update the title screen"""
    titleBlit() # Blit the moving background image
    titleBoxRect = titleBoxSurface.get_rect()
    screen.blit(titleBoxSurface,TITLE_BOX_POS)

    logoRect = titleScreenLogo.get_rect(centerx=SCREEN_CENTERX,centery=150)
    screen.blit(titleScreenLogo,logoRect)

    text = font.render(titleInfo[selectionID], 1, TEXT_COLOR)
    textRect = text.get_rect()
    titleInfoSurfaceTemp.fill(TRANS_COLOR)
    titleInfoSurfaceTemp.blit(titleInfoSurface,(0,0))
    titleInfoSurfaceTemp.blit(text,(UNIT_SIZE,TEXT_PADDING*1.3))
    titleInfoEndRect = (UNIT_SIZE+textRect[2],0,SCREEN_WIDTH,SCREEN_HEIGHT)
    titleInfoSurfaceTemp.fill(TRANS_COLOR,titleInfoEndRect)
    titleInfoSurfaceTemp.blit(titleInfoSurface2,(titleInfoEndRect[0],titleInfoEndRect[1]))
    
    infoBoxPos = (TITLE_BOX_POS[0]+titleBoxRect[2],TITLE_BOX_POS[1]+selectionID*UNIT_SIZE+TEXT_PADDING/2)
    screen.blit(titleInfoSurfaceTemp,infoBoxPos)
    
    titleContents(titleSections)
    showCursor(TITLE_TEXT_POS,selectionID)
    centeredText("Select with UP/DOWN keys, confirm with ENTER key",SCREEN_HEIGHT-FONT_SIZE,False,2)
    pygame.display.flip()

def titleScreenPanel(quantity,panelType):
    """Places a specified number of panels randomly on the title screen"""
    for i in range(quantity):
        pos = (SQUARE_SIZE*random.randint(0,int(TITLE_BG_WIDTH/SQUARE_SIZE)+1), \
               SQUARE_SIZE*random.randint(0,int(SCREEN_HEIGHT/SQUARE_SIZE)+1))
        titleBGSurface.blit(backgrounds,pos,(SQUARE_SIZE*panelType,0,SQUARE_SIZE,SQUARE_SIZE))

def titleScreenBG():
    """Generage the background image for the title screen"""
    for column in range(int(TITLE_BG_WIDTH/SQUARE_SIZE)+1):
        for row in range(int(SCREEN_HEIGHT/SQUARE_SIZE)+1):
            titleBGSurface.blit(backgrounds,(column*SQUARE_SIZE,row*SQUARE_SIZE), \
                    (0,0,SQUARE_SIZE,SQUARE_SIZE))

    titleScreenPanel(TITLE_FOREST_NUM,1)
    titleScreenPanel(TITLE_MOUNTAIN_NUM,2)
    titleScreenPanel(TITLE_BASE_NUM,3)
    titleScreenPanel(TITLE_BASE_NUM,4)

    for i in range(TITLE_UNIT_NUM):
        pos = (random.randint(UNIT_SIZE,TITLE_BG_WIDTH-UNIT_SIZE),random.randint(UNIT_SIZE,SCREEN_HEIGHT-UNIT_SIZE))
        titleBGSurface.blit(unitSurface,pos,(random.randint(0,5)*UNIT_SIZE,random.randint(0,3)*UNIT_SIZE,UNIT_SIZE,UNIT_SIZE))

def readFilter(returnInt=True):
    """Reads lines in a file without newlines, #'s, or tabs"""
    line = file.readline().split("\n")[0].split("#")[0].split("\t")[0]
    if returnInt:
        return int(line)
    else:
        return line

#################
# Load/Edit Functions
def fileLength(fileName):
    """Returns the number of lines in a file"""
    file = open(fileName,"r")
    
    totalLines = 0
    for line in file:
        totalLines += 1
    file.close()
    
    return totalLines+1

def reloadMap(empty=False):
    """Reload field and place units based on currently selected map"""
    Unit.clearUnits(Unit)
    Field.clear(Field)
    global fieldArray
    teamUnits = [list(),list()] # Initialize team unit list
    lineStart = mapNumber*10    # Start from this line (based on map #)

    if empty: # Generate a map full of grass
        fieldArray = [[0 for column in range(COLUMNS)] for row in range(ROWS)]
        field = Field(fieldArray)   # Generate a new field

    else:   # Generate the field
        file = open(MAPS_FILE,"r")
        fieldArray = []
        for line in range(lineStart):
            file.readline()     # Read lines until reaching the necessary line
        
        # Generate the field
        for row in range(ROWS):     # Read entries for all rows in the battlefield
            fieldArray.append(list(file.readline().split("\n")[0]))
            for column in range(COLUMNS): # Convert each row to a list of integers
                fieldArray[row][column] = int(fieldArray[row][column])
        field = Field(fieldArray)   # Generate a new field

        # Generate the units
        # Reads in the format: "unit type, x, y; unit type, x, y; unit type, x, y; etc"
        for team in range(2):
            teamUnits[team] = file.readline().split("\n")[0].split(";")
            if teamUnits[team] != "":
                for unit in range(len(teamUnits[team])-1):
                    unitInfo = teamUnits[team][unit].split(",")
                    UnitType(int(unitInfo[0]),team,(int(unitInfo[1]),int(unitInfo[2])))

        drawUnits()
        file.close()

def blitSelections():
    """Graphics for the unit/panel selection menu in the 'Edit Map' mode"""
    verticalOffset = TEXT_PADDING+BOUNDARY_BOTTOM
    horizontalOffset = TEXT_PADDING
    for panelType in range(5):  # Blit panels
        screen.blit(backgrounds2,(horizontalOffset,verticalOffset),(UNIT_SIZE*panelType,0,UNIT_SIZE,UNIT_SIZE))
        horizontalOffset += UNIT_SIZE*1.5
    for team in range(2):       # Blit units for each team
        verticalOffset += UNIT_SIZE*1.5
        horizontalOffset = TEXT_PADDING
        for unitType in range(6):
            screen.blit(unitSurface,(horizontalOffset,verticalOffset),(unitType*UNIT_SIZE,2*team*UNIT_SIZE,UNIT_SIZE,UNIT_SIZE))
            horizontalOffset += UNIT_SIZE*1.5
    screen.blit(cursorSurface,(TEXT_PADDING+UNIT_SIZE*1.5*editSelection,TEXT_PADDING+BOUNDARY_BOTTOM+UNIT_SIZE*1.5*editCategory),(UNIT_SIZE*3,0,UNIT_SIZE,UNIT_SIZE))

########################
# Help Screen Functions
def helpContents(strings):
    """Display the table of contents of the help menu on the left"""
    verticalOffset = HELP_CONTENTS_POS[1]
    for line in strings:
        text = font2.render(line, 1, TEXT_COLOR)
        screen.blit(text,(HELP_CONTENTS_POS[0],verticalOffset))
        verticalOffset += FONT_SIZE2

def helpMainText(strings):
    """Display the main text of the help menu"""
    verticalOffset = HELP_MAINTEXT_POS[1]
    for line in strings:
        imgType = None
        caption = False
        if len(line.split(".png")) > 1:     # Blitting images if line ends
            imgType = ".png"                #   in ".png" or ".jpg"
        elif len(line.split(".jpg")) > 1:
            imgType = ".jpg"
        if len(line.split("<c>")) > 1:      # Centered text if line begins
            caption = True                  #   with "<c>"
        
        if imgType != None:     # If the line is an image...
            # The entire image does not have to be blitted; parts can be chosen
            if line.split(imgType)[1] != "":
                selectiveBlitting = True    
                blitRect = line.split(imgType)[1].split(",")
                for i in range(4):
                    blitRect[i] = eval(blitRect[i])
            else:
                selectiveBlitting = False
                
            verticalOffset += TEXT_PADDING*.5

            # Make the image into a surface
            image = pygame.image.load(line.split(imgType)[0]+imgType).convert()
            image.set_colorkey(TRANS_COLOR)
            
            if selectiveBlitting:   # Part of the image is blitted
                tempSurface = pygame.Surface((blitRect[2],blitRect[3])).convert()
                imageRect = tempSurface.get_rect(top=verticalOffset,centerx=487)
                screen.blit(image,imageRect,blitRect)
            else:                   # The whole image is blitted
                imageRect = image.get_rect(top=verticalOffset,centerx=487)
                screen.blit(image,imageRect)
            verticalOffset += imageRect[3]
        else:
            if caption:     # Centered text, smaller font
                text = font_caption.render(line.split("<c>")[1], 1, TEXT_COLOR)
                textRect = text.get_rect(top=verticalOffset,centerx=487)
                screen.blit(text,textRect)
                verticalOffset += FONT_SIZE*.75
            else:       # Regular sized non-centered text
                text = font.render(line, 1, TEXT_COLOR)
                screen.blit(text,(HELP_MAINTEXT_POS[0],verticalOffset))
                verticalOffset += FONT_SIZE

def pageNumberText():
    """Show the current page number of the current help section"""
    text = "Page "+str(page+1)+" of "+str(len(helpText[sectionID]))
    pageText = font2.render(text, 1, TEXT_COLOR)
    textRect = pageText.get_rect(bottom=SCREEN_HEIGHT-TEXT_PADDING,centerx=487)
    screen.blit(pageText,textRect)
    if page > 0:
        screen.blit(cursorSurface,(textRect[0]-UNIT_SIZE*2,textRect[1]), (32,0,32,32))
    if page < len(helpText[sectionID])-1:
        screen.blit(cursorSurface,(textRect[0]+textRect[2]+UNIT_SIZE,textRect[1]), (64,0,32,32))

def helpScreenUpdate():
    """Update the help screen"""
    titleBlit()
    screen.blit(helpSurface,(0,0))
    helpContents(helpSections)
    helpMainText(helpText[sectionID][page])
    if len(helpText[sectionID]) > 1:
        pageNumberText()
    showCursor(HELP_CONTENTS_POS,sectionID)
    pygame.display.flip()

########################
# Simulation functions
def showText(strings,centered=False,offset=0):
    """Show text in the textbox. Input should be a tuple of strings"""
    verticalOffset = TEXT_PADDING+BOUNDARY_BOTTOM
    for line in strings:
        text = font.render(line, 1, TEXT_COLOR)
        if centered == False:
            textRect = (TEXT_PADDING+offset,verticalOffset)
        else:
            textRect = text.get_rect(top=verticalOffset,centerx=SCREEN_CENTERX)
        screen.blit(text,textRect)
        verticalOffset += FONT_SIZE

def showPausedText():
    """Show the word 'Paused' near the center of the screen"""
    pausedText = font.render("Paused", 1, TEXT_COLOR)
    pausedTextRect = pausedText.get_rect(centerx=BOUNDARY_RIGHT/2,centery=BOUNDARY_BOTTOM/2)
    screen.blit(pausedText, pausedTextRect)

def rectCoords(x1,y1,x2,y2):
    """Given coordinates of two points, returns a tuple for drawing a rectangle"""
    if x1 > x2:         # Make sure x2 is greater than x1; same with y
        x1,x2 = x2,x1
    if y1 > y2:
        y1,y2 = y2,y1
    rectWidth = x2-x1
    rectHeight = y2-y1
    return (x1,y1,rectWidth,rectHeight)

def coordinate(xTrue,yTrue):
    """Given coordinates on the screen, returns the position of the panel
        that corresponds to those coordinates"""
    xTrue = min(xTrue,SQUARE_SIZE*9)
    yTrue = min(yTrue,SQUARE_SIZE*5)
    return floor(xTrue/SQUARE_SIZE), floor(yTrue/SQUARE_SIZE)

def magnitude(xComp,yComp):
    """Returns the magnitude of a 2-dimensional vector"""
    return sqrt(xComp**2+yComp**2)

def speed(unit):
    """Returns the magnitude of the velocity vector of a unit"""
    return sqrt(unit.vel[0]**2+unit.vel[1]**2)

def distance(unit1,unit2):
    """Returns the distance between two units"""
    return sqrt((unit1.pos[0]-unit2.pos[0])**2+(unit1.pos[1]-unit2.pos[1])**2)

def tempDistance(unit1,unit2):
    """Returns the distance between two units (for use in RK4)"""
    return sqrt((unit1.tempPos[0]-unit2.tempPos[0])**2+(unit1.tempPos[1]-unit2.tempPos[1])**2)

def wpDistance(unit):
    """Returns the distance between a unit and its waypoint"""
    return sqrt((unit.tempPos[0]-unit.wpPos[0])**2+(unit.tempPos[1]-unit.wpPos[1])**2)

def acc(unit,comp):
    """Computes the acceleration of a unit in the specified component"""
    totalAcc = 0

    # Determine which units/bases are on the same team as the unit
    if unit.isFriendly:     
        enemyUnits = Unit.hostileList
        ownUnits = Unit.friendlyList
        enemyBases = Base.hostileList
        ownBases = Base.friendlyList
    else:
        enemyUnits = Unit.friendlyList
        ownUnits = Unit.hostileList
        enemyBases = Base.friendlyList
        ownBases = Base.hostileList
    
    if unit.hp > unit.tempEscape:   # Is the unit escaping?
        unit.escaping = -1
        if len(ownBases) > 0:       # Are there any same-colored bases?
            unit.tempEscape = unit.escape
    else:
        unit.escaping = 1
        if len(ownBases) > 0:
            unit.tempEscape = unit.maxHP*.9

    # Medics also escape when low on supplies
    if unit.type == 5 and unit.escaping == -1:
        if unit.supplies > unit.tempSupplyEscape:
            unit.escaping = -1
            unit.tempSupplyEscape = unit.supplyEscape
        else:
            unit.escaping = 1
            unit.tempSupplyEscape = unit.maxSupplies*.9

    # If the unit has a set waypoint, accelerate toward it
    if unit.waypointOn and unit.isFriendly and unit.escaping != 1:
        if unit.tempPos[comp] > unit.wpPos[comp]:
            multiplier = -1
        elif unit.tempPos[comp] < unit.wpPos[comp]:
            multiplier = 1
        else:
            multiplier = 0
        totalAcc += multiplier*(wpDistance(unit))/unit.hp

    # If escaping, accelerate toward the closest base
    if unit.escaping == 1:
        closestBase = None
        minDistance = 1e10
        for base in ownBases:
            currentDistance = distance(unit,base)
            if currentDistance < minDistance:
                minDistance = currentDistance
                closestBase = base

        # Determine which direction to accelerate toward
        if closestBase != None:
            if unit.tempPos[comp] > closestBase.pos[comp]:
                multiplier = -1
            elif unit.tempPos[comp] < closestBase.pos[comp]:
                multiplier = 1
            else:
                multiplier = 0
                
            totalAcc += multiplier*(distance(unit,closestBase)-SQUARE_SIZE/4)*unit.hp

    if unit.type == 5:
        # If the unit is a medic, move toward allies if not escaping
        for ownUnit in ownUnits:
            if unit.tempPos[comp] > ownUnit.tempPos[comp]:
                multiplier = -1
            else:
                multiplier = 1

            if ownUnit.hp > 1:
                totalAcc += multiplier*unit.hp*(tempDistance(unit,ownUnit))/(ownUnit.hp*4000)
        
    else:
        # Move toward enemies if not escaping; away from enemies if escaping
        for enemy in enemyUnits:
            # Determine the direction of acceleration
            if unit.tempPos[comp] < enemy.tempPos[comp]:
                multiplier = unit.escaping*(-1)
            else:
                multiplier = unit.escaping
                
            if enemy.hp > 1:
                totalAcc += multiplier*unit.hp*(tempDistance(unit,enemy)-unit.currentRange*.5)/(enemy.hp*5000)

    # If close to an enemy base, attempt to move away from it
    if unit.waypointOn == False:
        for base in enemyBases:
            if distance(unit,base) < 96:
                if unit.tempPos[comp] > base.pos[comp]:
                    multiplier = 1
                elif unit.tempPos[comp] < base.pos[comp]:
                    multiplier = -1
                else:
                    multiplier = 0
                totalAcc += multiplier*unit.hp/1000

    # Attempt to move away from mountains
    for mountain in Mountain.instanceList:
        if distance(unit,mountain) < SQUARE_SIZE:
            if unit.tempPos[comp] > mountain.pos[comp]:
                multiplier = 1
            elif unit.tempPos[comp] < mountain.pos[comp]:
                multiplier = -1
            else:
                multiplier = 0
            totalAcc += multiplier*unit.hp/1500
    
    return totalAcc
    
def RK4Acceleration():
    """Use RK4 to update the position and velocity of each unit on the battlefield"""
    for slopeNum in range(4):   # Calculate four RK4 slopes
        if slopeNum == 0:       # Offset position and velocity based on slope number
            offset = 0
        elif slopeNum == 1 or slopeNum == 2:
            offset = 0.5*STEP_SIZE
        elif slopeNum == 3:
            offset = STEP_SIZE

        for unit in Unit.instanceList:  # For every unit that is not dead, compute
            if unit.hp > 1:             #   position and velocity slopes
                for comp in range(2):
                    unit.tempPos[comp] = unit.pos[comp]+offset*unit.posSlope[slopeNum-1][comp]
                    unit.tempVel[comp] = unit.vel[comp]+offset*unit.velSlope[slopeNum-1][comp]
                    unit.posSlope[slopeNum][comp] = unit.vel[comp]+offset*unit.velSlope[slopeNum-1][comp]
                    unit.velSlope[slopeNum][comp] = acc(unit,comp)


    for unit in Unit.instanceList:  # For every unit that is not dead, compute the
        if unit.hp > 1:             #   final position and velocity after one step
            for comp in range(2):
                unit.vel[comp] += HCONST*(unit.velSlope[0][comp] \
                                          +2*unit.velSlope[1][comp] \
                                          +2*unit.velSlope[2][comp] \
                                          +unit.velSlope[3][comp])
                unit.speedCheck()   # Check if the unit is moving faster than its
                                    #   max speed; if it is, slow it down
                unit.pos[comp] += HCONST*(unit.posSlope[0][comp] \
                                          +2*unit.posSlope[1][comp] \
                                          +2*unit.posSlope[2][comp] \
                                          +unit.posSlope[3][comp])
                unit.positionCheck()    # Check to see if the unit has reached a
                                        #   boundary of the screen.
            

def healthRate(unit,slopeNum):
    """Calculate the rate at which a unit's health is gained/lost"""
    totalHealthRate = 0

    # Determine which units/bases are on the same team as the unit
    if unit.isFriendly:
        enemyUnits = Unit.hostileList
        ownUnits = Unit.friendlyList
        ownMedics = Unit.friendlyMedicList
        enemyBases = Base.hostileList
        ownBases = Base.friendlyList
    else:
        enemyUnits = Unit.friendlyList
        ownUnits = Unit.hostileList
        ownMedics = Unit.hostileMedicList
        enemyBases = Base.friendlyList
        ownBases = Base.hostileList

    # Calculate health loss based on enemy units within attacking range
    for enemy in enemyUnits:
        ownUnitsInRange = 0
        currentDistance = distance(unit,enemy)
        if enemy.hp > 1 and enemy.currentRange > currentDistance:            
            # How many allied units are nearby? Less damage is taken if there are
            #   more allied units nearby, as the damage is spread to all units in range
            for ownUnit in ownUnits:
                currentDistance2 = tempDistance(ownUnit,enemy)
                if currentDistance2 < enemy.currentRange:
                    ownUnitsInRange += 1
            if unit.currentPanel == SPEED_FOREST:
                # If the unit is in a forest, it takes less damage from enemies
                totalHealthRate -= sqrt(enemy.power*enemy.tempHP)/((5000+currentDistance)*max(1,ownUnitsInRange))
            else:
                # If the unit is not in a forest, it takes more damage in comparison
                totalHealthRate -= enemy.power*enemy.tempHP/((5000+currentDistance)*max(1,ownUnitsInRange))

    # Healing done by medics, only if enough supplies are available
    for medic in ownMedics:
        if unit.type != 5 and medic.hp > 1 and medic.supplies > 1:
            currentDistance = distance(unit,medic)
            if medic.currentRange > currentDistance and unit.tempHP < unit.maxHP:            
                medicHealRate = 100*medic.tempHP/(5000+currentDistance)
                totalHealthRate += medicHealRate
                medic.supplySlope[slopeNum] -= medicHealRate
                
    # If the unit is close enough to its own base, it gains health
    for base in ownBases:
        currentDistance = distance(unit,base)
        if currentDistance < SQUARE_SIZE:
            totalHealthRate += BASE_REINFORCEMENT*.2
            if unit.type == 5:  # If the unit is a medic, it gains supplies too
                unit.supplySlope[slopeNum] += BASE_REINFORCEMENT*.5

    # Take into consideration health rate based on reinforcements and accident rate
    if unit.currentPanel == SPEED_FOREST:
        # If the unit is in a forest, it takes more damage from accidents
        totalHealthRate += unit.reinforcements - 2*unit.tempHP*unit.accident
    else:
        totalHealthRate += unit.reinforcements - unit.tempHP*unit.accident
    
    return totalHealthRate

def RK4Health():
    """Uses RK4 to update the health of each unit on the battlefield"""
    for medic in Unit.medicList:
        for slopeNum in range(4):
            medic.supplySlope[slopeNum] = 0
    
    for slopeNum in range(4):   # Calculate four RK4 slopes
        if slopeNum == 0:       # Offset health based on slope number
            offset = 0
        elif slopeNum == 1 or slopeNum == 2:
            offset = 0.5*STEP_SIZE
        elif slopeNum == 3:
            offset = STEP_SIZE
        
        for unit in Unit.instanceList:  # For every unit that is not dead, calculate
            if unit.hp > 1:             #   the health slope of that unit for the
                                        #   current RK4 slope
                unit.tempHP = unit.hp+offset*unit.hpSlope[slopeNum-1]
                unit.hpSlope[slopeNum] = healthRate(unit,slopeNum)

    # Use a weighted average of all four RK4 slopes to update the health of units
    for unit in Unit.instanceList:
        if unit.hp >= 1:
            unit.hp += HCONST*(unit.hpSlope[0]+2*unit.hpSlope[1]+ \
                               2*unit.hpSlope[2]+unit.hpSlope[3])
            if unit.type == 5:  # Increase or decrease supplies for medic units
                unit.supplies += HCONST*(unit.supplySlope[0]+2*unit.supplySlope[1]+ \
                                   2*unit.supplySlope[2]+unit.supplySlope[3])
            unit.healthCheck()  # if unit health is greater than max health, adjust
        if unit.hp < 1: # If unit health is less than 1, the unit is dead
            unit.hp = 0

def resetSimulation():
    """Reset simulation variables"""
    t = 0
    clickPos = (0,0)
    leftClicking = False
    paused = False
    Field.clear(Field)

def drawUnits():
    """Based on each unit's properties, blit units to the screen"""
    if fow and onSimulation:
        fowSurface.fill(TEXT_COLOR)
    for unit in Unit.instanceList:
        if unit.hp > 1:
            # The following few lines draw the unit's vertical health bar
            # Based on how much percent health it has left
            healthFraction = unit.hp/unit.maxHP

            if unit.type < 2 and healthFraction > .5:
                unit.currentRange = unit.maxRange*healthFraction

            lineStart = (unit.pos[0]+UNIT_SIZE/2,unit.pos[1]+UNIT_SIZE/2)
            lineEnd1 = (unit.pos[0]+UNIT_SIZE/2,unit.pos[1]-UNIT_SIZE/2)
            lineEnd2 = (unit.pos[0]+UNIT_SIZE/2,unit.pos[1]+UNIT_SIZE*(.5-healthFraction))
        
            pygame.draw.line(screen, 0x000000, lineStart, lineEnd1, 4)
            pygame.draw.line(screen, 0xFFFFFF, lineStart, lineEnd1, 2)
            pygame.draw.line(screen, unit.color, lineStart, lineEnd2, 2)

            if unit.type == 5:  # If it's a medic unit, draw a supply bar also
                supplyFraction = unit.supplies/unit.maxSupplies
                
                lineStart = (unit.pos[0]+UNIT_SIZE/2+5,unit.pos[1]+UNIT_SIZE/2)
                lineEnd1 = (unit.pos[0]+UNIT_SIZE/2+5,unit.pos[1]-UNIT_SIZE/2)
                lineEnd2 = (unit.pos[0]+UNIT_SIZE/2+5,unit.pos[1]+UNIT_SIZE*(.5-supplyFraction))
            
                pygame.draw.line(screen, 0x000000, lineStart, lineEnd1, 4)
                pygame.draw.line(screen, 0xFFFFFF, lineStart, lineEnd1, 2)
                pygame.draw.line(screen, SUPPLY_COLOR, lineStart, lineEnd2, 2)
            
            if unit.vel[0] > 0: # The direction the sprite faces depends on velocity
                facingRight = 1
            else:
                facingRight = 0

            # Blit the unit's graphic to the screen    
            screen.blit(unitSurface,(unit.pos[0]-UNIT_SIZE/2,unit.pos[1]-UNIT_SIZE/2),(unit.type*UNIT_SIZE,(unit.isFriendly*2+facingRight)*UNIT_SIZE,UNIT_SIZE,UNIT_SIZE))

            # Draw circles representing range
            if rangeEnabled and not unit.isSelected:
                pygame.draw.circle(screen, unit.color, (int(unit.pos[0]),int(unit.pos[1])), int(unit.currentRange),1)
            if unit.isSelected:
                pygame.draw.circle(screen, unit.color, (int(unit.pos[0]),int(unit.pos[1])), int(unit.currentRange),2)
            if fow and unit.isFriendly and onSimulation:
                pygame.draw.circle(fowSurface, TRANS_COLOR, (int(unit.pos[0]),int(unit.pos[1])), int(unit.currentRange),0)
            # Draw waypoint and line connecting unit and waypoint
            if unit.waypointOn and unit.escaping != 1:
                screen.blit(wpSurface,wpSurface.get_rect(center=unit.wpPos))
                pygame.draw.line(screen,FRIENDLY_COLOR,unit.pos,unit.wpPos)
        else:
            screen.blit(graveSurface,(unit.pos[0]-UNIT_SIZE/4,unit.pos[1]-UNIT_SIZE/4),(unit.isFriendly*UNIT_SIZE/2,0,UNIT_SIZE/2,UNIT_SIZE/2))
    if fow and onSimulation:
        screen.blit(fowSurface,(0,0))

#########################
# Pygame Initialization #
#########################

# Initialize pygame window for display
pygame.init()
iconSurface = pygame.image.load("images/icon.png") # Load icon image
pygame.display.set_icon(iconSurface)        # Set the window's icon
screen = pygame.display.set_mode((SCREEN_WIDTH,SCREEN_HEIGHT))
pygame.display.set_caption(TITLE)
screen.fill(BG_COLOR)

# Set different font sizes
font = pygame.font.Font(pygame.font.match_font(FONT_NAME), FONT_SIZE)
font2 = pygame.font.Font(pygame.font.match_font(FONT_NAME), FONT_SIZE2)
font_caption = pygame.font.Font(pygame.font.match_font(FONT_NAME), int(FONT_SIZE*.75))


#######################
# Initialize Surfaces #
#######################

# Main selection box on title screen
titleBoxSurface = pygame.image.load("images/selection.png").convert()
titleBoxSurface.set_colorkey(TRANS_COLOR)

# Title screen logo
titleScreenLogo = pygame.image.load("images/logo.png").convert()
titleScreenLogo.set_colorkey((136,136,136))

# Info box on title screen
titleInfoSurface = pygame.image.load("images/explanation.png").convert()
titleInfoSurface2 = pygame.image.load("images/explanation2.png").convert()
titleInfoSurface.set_colorkey(TRANS_COLOR)
titleInfoSurface2.set_colorkey(TRANS_COLOR)
titleInfoSurfaceTemp = pygame.Surface(titleInfoSurface.get_size())
titleInfoSurfaceTemp.set_colorkey(TRANS_COLOR)

# Selection cursor
cursorSurface = pygame.image.load("images/cursors.png").convert()
cursorSurface.set_colorkey(TRANS_COLOR)

# Title background surface
titleBGSurface = pygame.Surface((TITLE_BG_WIDTH,SCREEN_HEIGHT))

# Battlefield panels
backgrounds = pygame.image.load("images/panels.png").convert()
backgrounds2 = pygame.transform.scale(backgrounds,(int(backgrounds.get_width()/2),int(backgrounds.get_height()/2)))

# Help screen
helpSurface = pygame.image.load("images/help.png").convert()
helpSurface.set_colorkey(TRANS_COLOR)

# Create a surface for unit graphics
unitSurface = pygame.image.load("images/units.png").convert()
unitSurface.set_colorkey(TRANS_COLOR)

# Create a surface for the waypoint graphic, which is a circle with an x through it
wpSurface = pygame.Surface((WP_RADIUS*2,WP_RADIUS*2))
wpSurface.fill(TRANS_COLOR)
wpSurface.set_colorkey(TRANS_COLOR)
pygame.draw.line(wpSurface,WP_COLOR,(0,0),(WP_RADIUS*2,WP_RADIUS*2),2)
pygame.draw.line(wpSurface,WP_COLOR,(WP_RADIUS*2,0),(0,WP_RADIUS*2),2)
pygame.draw.circle(wpSurface,WP_COLOR,(WP_RADIUS,WP_RADIUS),WP_RADIUS,2)

# Create a surface for the background (battlefield grid)
bgSurface = pygame.Surface((BOUNDARY_RIGHT,BOUNDARY_BOTTOM))

# Fog of war surface
fowSurface = pygame.Surface((BOUNDARY_RIGHT,BOUNDARY_BOTTOM))
fowSurface.set_colorkey(TRANS_COLOR)

# Create a surface for a textbox
textboxSurface = pygame.image.load("images/textbox.png").convert()
textboxPos = (0,BOUNDARY_BOTTOM)

# Create a surface for gravestone graphics
graveSurface = pygame.image.load("images/gravestones.png").convert()
graveSurface.set_colorkey(TRANS_COLOR)

#####################
# Class Definitions #
#####################

class Field(object):
    """The battlefield is split up into a 10x6 grid, with each element in
        the grid posessing properties depending on its type."""
    bases = []
    mountains = []
    def __init__(self,array):
        self.element = array
        backgrounds = pygame.image.load("images/panels.png").convert() # Load bg images

        # Use the appropriate image for each panel in the battlefield based on type
        for row in range(ROWS):
            for column in range(COLUMNS):
                bgSurface.blit(backgrounds,(column*SQUARE_SIZE,row*SQUARE_SIZE), \
                    (self.element[row][column]*SQUARE_SIZE,0,SQUARE_SIZE,SQUARE_SIZE))

                pos = (SQUARE_SIZE*(column+.5),SQUARE_SIZE*(row+.5))

                # If the panel is a mountain, create a new mountain object instance
                if self.element[row][column] == 2:
                    self.mountains.append(Mountain(pos))

                # If the panel is a base, create a new base object instance
                if self.element[row][column] == 3 or self.element[row][column] == 4:
                    baseType = self.element[row][column] - 3
                    self.bases.append(Base(baseType,pos))

        screen.blit(bgSurface,(0,0))    # Blit the completed background to the screen

    def panelType(self,xTrue,yTrue):
        """Given coordinates on the screen, return the type of the panel located
            at that coordinate"""
        xCoord,yCoord = coordinate(xTrue,yTrue)
        return self.element[yCoord][xCoord]

    def clear(self):
        """Clear all arrays associated with the field"""
        bases = []
        mountains = []
        Base.instanceList = []
        Base.friendlyList = []
        Base.hostileList = []
        Mountain.instanceList = []

class Mountain(object):
    """Keeps track of the coordinates of each mountain"""
    instanceList = []
    def __init__(self,pos):
        self.instanceList.append(self)
        self.pos = [pos[0],pos[1]]

class Base(object):
    """Keeps track of the coordinates and team of each base"""
    instanceList = []
    friendlyList = []
    hostileList = []
    def __init__(self,isFriendly,pos):
        self.instanceList.append(self)
        self.isFriendly = isFriendly
        if isFriendly:
            self.friendlyList.append(self)
            self.color = FRIENDLY_COLOR
        else:
            self.hostileList.append(self)
            self.color = HOSTILE_COLOR
        self.pos = [pos[0],pos[1]]

class Unit(object):
    """A unit on the battlefield; it may be friendly (blue) or hostile (red)"""
    instanceList = []
    friendlyList = []
    friendlyMedicList = []
    hostileList = []
    hostileMedicList = []
    medicList = []
    selectedList = []
    def __init__(self,isFriendly,pos,hp,speed,power,maxRange,accident,escape,reinforcements):
        """Initialize a unit based on various attributes"""
        self.instanceList.append(self)  # Append this unit to the list of all units
        self.isFriendly = isFriendly
        self.isSelected = False
        self.waypointOn = False

        # Determine which team the unit is on, and append it to the appropriate list
        if isFriendly:
            self.friendlyList.append(self)
            self.color = FRIENDLY_COLOR
        else:
            self.hostileList.append(self)
            self.color = HOSTILE_COLOR
            
        self.pos = [pos[0],pos[1]]      # Initial x and y coordinates of the unit
        self.tempPos = [pos[0],pos[1]]
        self.vel = [0,0]                # Units start at rest
        self.tempVel = [0,0]
        self.hp = self.maxHP = hp       # Initial and max HP (health points) of the unit
        self.tempHP = hp
        self.maxSpeed = speed           # Maximum speed allowed for the unit
        self.power = power              # Unit power; a higher value results in more damage
        self.currentRange = self.maxRange = maxRange        # Maximum attacking range of the unit
        self.accident = accident        # A percentage of unit's health is lost due to accidents
        self.escape = self.tempEscape = escape*hp         # Escape health; if HP falls below this, it attempts to flee
        self.reinforcements = reinforcements    # Constant reinforcement rate
        self.escaping = False

        # Initialize slopes used in RK4
        self.posSlope = [[0.0 for comp in range(2)] for slopeNum in range(4)]
        self.velSlope = [[0.0 for comp in range(2)] for slopeNum in range(4)]
        self.hpSlope = [0.0 for slopeNum in range(4)]

    def speedRescale(self,multiplier):
        """Multiplies the x and y components of velocity by a specified value"""
        self.vel[0] *= multiplier
        self.vel[1] *= multiplier

    def speedCheck(self):
        """If the unit's current speed exceeds its maximum speed, slow it down"""
        self.currentSpeed = speed(self)
        maxSpeed = self.maxSpeed*self.currentPanel()
        if self.currentSpeed > maxSpeed:
            speedScale = maxSpeed/self.currentSpeed
            self.speedRescale(speedScale)
        self.RKSpeedCheck()

    def RKSpeedCheck(self):
        """Rescale position slopes used in RK4 if they exceed unit's maximum speed"""
        for slopeNum in range(4):
            RKSpeed = magnitude(self.posSlope[slopeNum][0],self.posSlope[slopeNum][1])
            maxSpeed = self.maxSpeed*self.currentPanel()
            if RKSpeed > maxSpeed:
                speedScale = maxSpeed/RKSpeed
                self.posSlope[slopeNum][0] *= speedScale
                self.posSlope[slopeNum][1] *= speedScale

    def positionCheck(self):
        """If the unit hits a boundary of the screen, reverse its velocity"""
        if self.pos[0] < 0:
            self.pos[0] = 0
            self.vel[0] *= -1
        elif self.pos[0] > BOUNDARY_RIGHT:
            self.pos[0] = BOUNDARY_RIGHT
            self.vel[0] *= -1
        if self.pos[1] < 0:
            self.pos[1] = 0
            self.vel[1] *= -1
        elif self.pos[1] > BOUNDARY_BOTTOM:
            self.pos[1] = BOUNDARY_BOTTOM
            self.vel[1] *= -1

    def healthCheck(self):
        """If unit HP exceeds its max HP, lower it"""
        if self.hp > self.maxHP:
            self.hp = self.maxHP
        if self.type == 5:
            if self.supplies > self.maxSupplies:
                self.supplies = self.maxSupplies
            if self.supplies < 1:
                self.supplies = 0

    def currentPanel(self):
        """Returns the speed multiplier of the panel the unit is currently on"""
        panel = field.panelType(self.pos[0],self.pos[1])
        if panel == 1:
            return SPEED_FOREST
        elif panel == 2:
            return SPEED_MOUNTAIN
        else:
            return SPEED_NORMAL

    def waypoint(self,on,coords=None):
        """Sets a waypoint coordinate for the unit; 'on' should be True/False"""
        self.waypointOn = on
        if on == True:
            self.wpPos = coords

    def inRect(self):
        """If the unit is in the selection rectangle, return True"""
        for comp in range(2):
            if mousePos[comp]+UNIT_SIZE/2 > self.pos[comp] > clickPos[comp]-UNIT_SIZE/2 or \
               mousePos[comp]-UNIT_SIZE/2 < self.pos[comp] < clickPos[comp]+UNIT_SIZE/2:
                continue
            else:
                return False
        return True

    def remove(self):
        """Remove a unit from the lists referencing it"""
        self.instanceList.remove(self)
        if self.isFriendly:
            self.friendlyList.remove(self)
            if self.type == 5:
                self.friendlyMedicList.remove(self)
                self.medicList.remove(self)
        else:
            self.hostileList.remove(self)
            if self.type == 5:
                self.hostileMedicList.remove(self)
                self.medicList.remove(self)

    def clearUnits(self):
        self.instanceList = []
        self.friendlyList = []
        self.friendlyMedicList = []
        self.hostileList = []
        self.hostileMedicList = []
        self.medicList = []
        self.selectedList = []

class Medic(Unit):
    """A medic unit, carries supplies used to heal injured teammates"""
    def __init__(self,isFriendly,pos):
        self.isSelected = False
        self.waypointOn = False
        
        self.isFriendly = isFriendly

        # Add self to the appropriate lists
        Unit.medicList.append(self)
        Unit.instanceList.append(self)
        if isFriendly:
            Unit.friendlyList.append(self)
            Unit.friendlyMedicList.append(self)
            self.color = FRIENDLY_COLOR
        else:
            Unit.hostileList.append(self)
            Unit.hostileMedicList.append(self)
            self.color = HOSTILE_COLOR

        self.supplies = self.maxSupplies = 700
        self.pos = [pos[0],pos[1]]      # Initial x and y coordinates of the unit
        self.tempPos = [pos[0],pos[1]]
        self.vel = [0,0]                # Units start at rest
        self.tempVel = [0,0]
        self.hp = self.tempHP = self.maxHP = 200      # Initial and max HP (health points) of the unit
        self.maxSpeed = 2.5             # Maximum speed allowed for the unit
        self.power = 0                  # No power; this unit does not attack.
        self.currentRange = self.maxRange = 100             # Maximum healing range of the unit
        self.accident = .0015           # A percentage of unit's health is lost due to accidents
        self.escape = self.tempEscape = .3         # Escape health; if HP falls below this, it attempts to flee
        self.supplyEscape = self.tempSupplyEscape = self.maxSupplies*.2
        self.reinforcements = .5         # Constant reinforcement rate
        self.escaping = False

        # Initialize slopes used in RK4
        self.posSlope = [[0.0 for comp in range(2)] for slopeNum in range(4)]
        self.velSlope = [[0.0 for comp in range(2)] for slopeNum in range(4)]
        self.hpSlope = [0.0 for slopeNum in range(4)]
        self.supplySlope = [0.0 for slopeNum in range(4)]

class UnitType(object):
    def __init__(self,unitType,team,pos):
        """Creates a new unit with pre-specified values"""
        if unitType <= 0:    # Average unit
            newUnit = Unit(team,pos,325,3,100,75,.003,.4,1)
        elif unitType == 1:  # Slower but more powerful version of average unit
            newUnit = Unit(team,pos,325,2,130,75,.004,.4,1)
        elif unitType == 2:  # Fast unit with low health but long range
            newUnit = Unit(team,pos,280,4,100,115,.004,.35,.75)
        elif unitType == 3:  # Slow unit with high health and power
            newUnit = Unit(team,pos,480,1.5,90,50,.002,.4,1.25)
        elif unitType == 4:  # Very slow long range unit; does not flee
            newUnit = Unit(team,pos,350,.5,100,125,.001,0,.5)
        elif unitType == 5:  # Medic; heals allied units
            newUnit = Medic(team,pos)
        Unit.instanceList[len(Unit.instanceList)-1].type = unitType

########################
# Initialize variables #
########################

# Title screen variables
selectionID = 0
titleOffset = 0

# Variables representing current mode
onTitleScreen = True
onRandomization = onLoad = onHelpScreen = onSimulation = onEdit = editingMap = False

# Map loading variables
totalMaps = math.floor(fileLength(MAPS_FILE)/10) # Number of maps saved
mapNumber = 0   # Current selected map
fieldArray = list()
teamUnits = [list(),list()] # Initialize team unit list

# Simulation variables
mousePos = (0,0)
clickPos = (0,0)
rangeEnabled = False
leftClicking = False
rightClicking = False
paused = False
fow = False             # Fog of war
Field.clear(Field)

########
# Main #
########

titleScreenBG()
titleScreenUpdate()
while True:
    if onTitleScreen:
        titleOffset -= TITLE_SCROLL_SPEED   # Background scrolls to the left
        if titleOffset < -TITLE_BG_WIDTH:
            titleOffset = 0
        titleScreenUpdate()     # Update title screen
        Unit.clearUnits(Unit)   # Clear all units
        
        for event in pygame.event.get():        
            if event.type == QUIT:
                os._exit(99)
            if event.type == KEYDOWN:
                if event.key == K_ESCAPE:   # Esc closes the game
                    os._exit(99)
                if event.key == K_UP:       # Arrow keys to move between options
                    if selectionID > 0: 
                        selectionID -= 1
                    else:
                        selectionID = len(titleSections)-1
                if event.key == K_DOWN:
                    if selectionID < len(titleSections)-1:
                        selectionID += 1
                    else:
                        selectionID = 0
                if event.key == K_RETURN:   # Enter key to confirm selection
                    if selectionID == 0:    # Does different things depending on what is selected
                        onRandomization = True
                    if selectionID == 1:
                        onLoad = True
                        waitingForInput = False
                    if selectionID == 2:
                        onEdit = True
                        selectingMap = True
                        waitingForInput = False
                    if selectionID == 3:
                        onHelpScreen = True
                    if selectionID == 5:
                        os._exit(99)
                    onTitleScreen == False  # Exit this loop
                    
    if onRandomization:     # Randomize a map based on settings (from a file)
        file = open(SETTINGS_FILE,"r")
        file.readline()     # First line is filler text

        # Begin loading info from the file
        numForests = readFilter()   
        numMountains = readFilter()
        numBases = [readFilter(),readFilter()]  # Number of red/blue bases
        numUnits = [readFilter(),readFilter()]  # Number of red/blue units
        startNearBase = [readFilter(),readFilter()] # Red/blue starts near base? True/False
        baseLocation = [list(),list()]          # Initialize arrays for base locations
        teamCoords = [readFilter(False).split(","),readFilter(False).split(",")] # If startNearBase is false, these coordinates specify where they start
        teamMakeup = [readFilter(False).split(","),readFilter(False).split(",")] # What is each team made up of, unit-wise? (In terms of a percent for each unit type)

        fieldArray = [[0 for column in range(COLUMNS)] for row in range(ROWS)] # Every panel starts out as a grass
        panelPos = (random.randint(0,ROWS-1),random.randint(0,COLUMNS-1)) # Choose a random panel
        for panelType in reversed(range(4)):    # Start randomly placing bases, then mountains, then forests
            whichNum = (numForests,numMountains,numBases[0],numBases[1])[panelType] # Which variable to use depends on panel type
            for panel in range(whichNum):
                while fieldArray[panelPos[0]][panelPos[1]] != 0:    # If the panel is not grass, find a new panel
                    panelPos = (random.randint(0,ROWS-1),random.randint(0,COLUMNS-1))
                fieldArray[panelPos[0]][panelPos[1]] = panelType+1  # Set that panel to the current panel type

                if panelType == 2 or panelType == 3 and startNearBase[panelType-3] == True: # If the panel is a base, add it to the list
                    baseLocation[panelType-2].append((SQUARE_SIZE*(panelPos[0]+.5),SQUARE_SIZE*(panelPos[1]+.5)))
                    
        teamUnits = [list(),list()]     # Initialize unit list
        for team in range(2):
            for unitType in range(6):
                teamUnits[team].append(math.floor(float(teamMakeup[team][unitType])*numUnits[team])) # Determine the number of units on the team based on total number of units
                if float(teamMakeup[team][unitType]) > 0 and teamUnits[team][unitType] == 0:
                    teamUnits[team][unitType] = 1 # If the team makeup has a non-zero value for this unit type, add a unit
              
            extraUnits = 0
            while sum(teamUnits[team]) < numUnits[team]:    # If not enough units, add a new unit starting from unit type 0 and moving up
                teamUnits[team][extraUnits%6] += 1
                extraUnits += 1

            subtractedUnits = 0
            while sum(teamUnits[team]) > numUnits[team]:    # If too many units, remove units starting from unit type 5 and moving down
                teamUnits[team][5-subtractedUnits%6] -= 1
                subtractedUnits += 1
                
            if not startNearBase[team]:     # If 'start near base' is false
                for rectCoord in range(4):  # Set starting coordinate range for the units
                    teamCoords[team][rectCoord] = int(teamCoords[team][rectCoord])

            for unit in range(6): 
                for i in range(teamUnits[team][unit]):
                    if startNearBase[team]: # Choose a random base for each unit to start near
                        whichBase = random.randint(0,len(baseLocation[team])-1)
                        xCoord = baseLocation[team][whichBase][1]+random.uniform(-DIST_FROM_BASE,DIST_FROM_BASE)
                        yCoord = baseLocation[team][whichBase][0]+random.uniform(-DIST_FROM_BASE,DIST_FROM_BASE)
                        UnitType(unit,team,(xCoord,yCoord))
                    else:   # Choose a coordinate in the rectangle to place a unit
                        xCoord = random.randint(teamCoords[team][0],teamCoords[team][2])
                        yCoord = random.randint(teamCoords[team][1],teamCoords[team][3])
                        UnitType(unit,team,(xCoord,yCoord))

        onRandomization = False # End randomization
        onSimulation = True     # Begin simulation

    while onEdit:   # Map editing mode
        mapNumber = totalMaps
        editCategory = 0
        editSelection = 0
        fileLines = []
        mapSaved = False
        textDisappearTime = 0

        file = open(MAPS_FILE,"r")  # Append file lines to a list
        for line in file:
            fileLines.append(line)

        # If file does not end in a newline character, add one
        if len(fileLines[len(fileLines)-1].split("\n")) == 1:
            fileLines[len(fileLines)-1] += "\n"

        # Map selection
        while selectingMap and editingMap == False:
            if not waitingForInput:
                reloadMap(mapNumber == totalMaps)
                screen.blit(textboxSurface,textboxPos)
                showText(("Select a map to edit with the LEFT and RIGHT arrow keys",
                          "Confirm selection with the ENTER key",
                          "Return to main menu with the ESC key"),True)
                if mapNumber == totalMaps:
                    centeredText("New map",550,True)
                else:
                    centeredText("Map %d of %d" % (mapNumber+1,totalMaps),550,True)
                screen.blit(cursorSurface,(TEXT_PADDING,TEXTBOX_CENTERY), (32,0,32,32))
                screen.blit(cursorSurface,(SCREEN_WIDTH-TEXT_PADDING-UNIT_SIZE,TEXTBOX_CENTERY), (64,0,32,32))
                
                pygame.display.flip()
                waitingForInput = True

            while waitingForInput:
                for event in pygame.event.get():        
                    if event.type == QUIT:
                        os._exit(99)
                    if event.type == KEYDOWN:
                        waitingForInput = False
                        if event.key == K_ESCAPE:   # Esc returns to main menu
                            onEdit = False
                            onTitleScreen = True
                        if event.key == K_LEFT:       # Arrow keys to move between options
                            if mapNumber > 0: 
                                mapNumber -= 1
                            else:
                                mapNumber = totalMaps
                        if event.key == K_RIGHT:
                            if mapNumber < totalMaps:
                                mapNumber += 1
                            else:
                                mapNumber = 0

                        if event.key == K_SPACE:    # Spacebar toggles max range circles
                            rangeEnabled = not rangeEnabled
                                
                        if event.key == K_RETURN:   # Enter key to confirm selection
                            selecingMap = False          # Stop loading
                            editingMap = True
                        if event.key == K_ESCAPE:
                            onEdit = False
                            editingMap = False
                            onTitleScreen = True
                            selectingMap = False
                            
        # Map editing
        while editingMap:            
            mousePos = pygame.mouse.get_pos()   # Get current mouse position
            mousePos2 = (math.floor(mousePos[0]/SQUARE_SIZE),math.floor(mousePos[1]/SQUARE_SIZE)) # Current panel position

            screen.blit(bgSurface,(0,0))
            drawUnits()

            # Set panels (The mouse button can be held down)
            if leftClicking and mousePos[1] < BOUNDARY_BOTTOM and editCategory == 0:
                bgSurface.blit(backgrounds,(mousePos2[0]*SQUARE_SIZE,mousePos2[1]*SQUARE_SIZE), \
                    (editSelection*SQUARE_SIZE,0,SQUARE_SIZE,SQUARE_SIZE))
                fieldArray[mousePos2[1]][mousePos2[0]] =  editSelection

            # Draw rectangle for deleting units
            if rightClicking:
                pygame.draw.rect(screen,TEXT_COLOR,rectCoords(clickPos[0],clickPos[1],mousePos[0],mousePos[1]),1)

            screen.blit(textboxSurface,textboxPos)
            showText(("Arrow keys to select panel/unit type",
                      "LClick to set panel/unit / RClick to delete units",
                      "ESC to exit to main menu / S to save map",
                      "ENTER to run simulation without saving","",
                      "Red units: %d / Blue units: %d / Mouse: (%d,%d)" % (len(Unit.hostileList),len(Unit.friendlyList),mousePos[0],mousePos[1])
                      ),offset=300)
            blitSelections()    # Unit/panel selection menu graphics

            # When a map is saved, show a message for some amount of time
            if mapSaved == True and textDisappearTime < 125:
                textDisappearTime += 1
                centeredText("Map saved to slot %d" % (mapNumber+1),BOUNDARY_BOTTOM/2,True,3)
            
            pygame.display.flip()
            
            for event in pygame.event.get():        
                if event.type == QUIT:
                    os._exit(99)

                if event.type == MOUSEBUTTONDOWN and mousePos[1] < BOUNDARY_BOTTOM:
                    if event.button == 1:   # Left mouse button pressed
                        if editCategory == 0:
                            leftClicking = True
                        else:
                            UnitType(editSelection,editCategory-1,mousePos)

                    if event.button == 3:   # Right mouse button pressed
                        clickPos = mousePos
                        rightClicking = True

                if event.type == MOUSEBUTTONUP:
                    if event.button == 1:
                        leftClicking = False
                    if event.button == 3:
                        rightClicking = False
                        for unit in Unit.instanceList:  # All units in selection rectangle
                            if unit.inRect():           #   are removed
                                unit.remove()       
                if event.type == KEYDOWN:
                    if event.key == K_ESCAPE:   # Esc returns to main menu
                        onEdit = False
                        onTitleScreen = True
                        
                    # Arrow keys to select units/panels
                    if event.key == K_RIGHT:
                        if editSelection == (4,5,5)[editCategory]:
                            editSelection = 0
                        else:
                            editSelection += 1
                            
                    if event.key == K_LEFT:
                        if editSelection == 0:
                            editSelection = (4,5,5)[editCategory]
                        else:
                            editSelection -= 1
                            
                    if event.key == K_UP:
                        if editCategory == 0:
                            editCategory = 2
                        else:
                            editCategory -= 1
                            
                    if event.key == K_DOWN:
                        if editCategory == 2:
                            editCategory = 0
                        else:
                            editCategory += 1
                            
                    if event.key == K_RETURN:   # Enter key to confirm selection
                        onEdit = False
                        editingMap = False
                        onSimulation = True
                        
                    if event.key == K_ESCAPE:   # ESC to return to title
                        onEdit = False
                        editingMap = False
                        onTitleScreen = True

                    if event.key == K_SPACE:    # Spacebar toggles max range circles
                        rangeEnabled = not rangeEnabled

                    if event.key == K_s:    # S key to save map
                        lineOffset = 0

                        # Write map panels to text file
                        for row in range(ROWS):
                            rowString = ""
                            for column in range(COLUMNS):
                                rowString += str(fieldArray[row][column])
                            rowString += "\n"
                            if mapNumber == totalMaps:
                                fileLines.append(rowString)
                            else:
                                fileLines[mapNumber*10+lineOffset] = rowString
                                lineOffset += 1

                        # Write units to text file
                        unitStringArray = ["",""]
                        for unit in Unit.instanceList:
                            unitStringArray[unit.isFriendly] += "%d,%d,%d;" % (unit.type,unit.pos[0],unit.pos[1])
                        if mapNumber == totalMaps:
                            fileLines.append(unitStringArray[0]+"\n")
                            fileLines.append(unitStringArray[1]+"\n")
                        else:
                            for team in range(2):
                                fileLines[mapNumber*10+lineOffset] = unitStringArray[team]+"\n"
                                lineOffset += 1
                        
                        file = open(MAPS_FILE,"w")
                        for line in fileLines:
                            file.write(line)
                        
                        if mapNumber == totalMaps:
                            file.write("-\n")
                            totalMaps += 1

                        file.close()
                        
                        textDisappearTime = 0
                        mapSaved = True
                        
                    if editCategory == 0 and editSelection == 5:
                        editSelection = 4
                        
    while onLoad:  # Load a map from file
        Unit.clearUnits(Unit)
        file = open(MAPS_FILE,"r")
        if mapNumber == totalMaps:
            mapNumber = 0

        if not waitingForInput:
            reloadMap()
            
            screen.blit(textboxSurface,textboxPos)
            showText(("Select map with the LEFT and RIGHT arrow keys",
                      "Begin simulation with the ENTER key",
                      "Return to main menu with the ESC key"),True)
            centeredText("Map %d of %d" % (mapNumber+1,totalMaps),550,True)
            screen.blit(cursorSurface,(TEXT_PADDING,TEXTBOX_CENTERY), (32,0,32,32))
            screen.blit(cursorSurface,(SCREEN_WIDTH-TEXT_PADDING-UNIT_SIZE,TEXTBOX_CENTERY), (64,0,32,32))
            pygame.display.flip()

            file.close()
            waitingForInput = True

        if totalMaps == 1:  # If there's only one saved map, start simulation
            onLoad = False          # Stop loading
            onSimulation = True     # Begin simulation
            waitingForInput = False

        while waitingForInput:  # If more than one map, the user chooses which one
            for event in pygame.event.get():        
                if event.type == QUIT:
                    os._exit(99)
                if event.type == KEYDOWN:
                    if event.key == K_ESCAPE:   # Esc returns to main menu
                        onLoad = False
                        onTitleScreen = True
                    if event.key == K_LEFT:       # Arrow keys to move between options
                        if mapNumber > 0: 
                            mapNumber -= 1
                        else:
                            mapNumber = totalMaps-1
                    if event.key == K_RIGHT:
                        if mapNumber < totalMaps-1:
                            mapNumber += 1
                        else:
                            mapNumber = 0

                    if event.key == K_SPACE:    # Spacebar toggles max range circles
                        rangeEnabled = not rangeEnabled
                        
                    if event.key == K_RETURN:   # Enter key to confirm selection
                        onLoad = False          # Stop loading
                        onSimulation = True     # Begin simulation
                    waitingForInput = False
        
    if onSimulation:
        resetSimulation()           # Reset variables
        field = Field(fieldArray)   # Generate the field
        mins = 0
        hours = 0
        days = 0
    ### Start of simulation loop
        while onSimulation:
            if not paused:
                mins += STEP_SIZE                      # Increment time
                if mins > 60:
                    mins -= 60
                    hours += 1
                if hours > 24:
                    hours -= 24
                    days += 1
                
                # Update health, position, and velocity with RK4
                RK4Health()
                RK4Acceleration()

            mousePos = pygame.mouse.get_pos()   # Get current mouse position
            
            screen.fill(BG_COLOR)
            screen.blit(bgSurface,(0,0))    # Blit the background to the screen

            # Draw units, health bar, their attack ranges, and waypoints
            drawUnits()

            # If the player has the left mouse button down, draw a selection rectangle
            if leftClicking:
                pygame.draw.rect(screen,FRIENDLY_COLOR,rectCoords(clickPos[0],clickPos[1],mousePos[0],mousePos[1]),1)

            screen.blit(textboxSurface,textboxPos)
            showText(("Left click + drag to select blue units",
                      "Right click to create waypoints for selected units",
                      "A to select all units / Z to clear all waypoints / P to pause",
                      "SPACE to show attacking range of all units","",
                      "Time elapsed: %d days, %d hours, %d minutes" % (days, hours, mins)))
            if paused:
                centeredText("Paused",BOUNDARY_BOTTOM/2,True,3)

            pygame.display.flip()

            # Check to see which keys or mouse buttons are pressed
            for event in pygame.event.get():        
                if event.type == QUIT:
                    os._exit(99)
                if event.type == MOUSEBUTTONDOWN:
                    if mousePos[1] < BOUNDARY_BOTTOM:
                        clickPos = mousePos
                        if event.button == 1:   # Left mouse button pressed
                            Unit.selectedList = []
                            for unit in Unit.friendlyList:  # De-select all units
                                unit.isSelected = False
                                unit.color = FRIENDLY_COLOR
                            leftClicking = True
                        if event.button == 3:   # Right mouse button pressed
                            for unit in Unit.selectedList:  # Create waypoints
                                unit.waypoint(True,clickPos)
                if event.type == MOUSEBUTTONUP:
                    if event.button == 1:       # Left mouse button released
                        leftClicking = False
                        for unit in Unit.friendlyList:  # All units in selection rectangle
                            if unit.inRect():           #   are added to a list
                                Unit.selectedList.append(unit)
                                unit.isSelected = True
                                unit.color = FRIENDLY_COLOR2                    
                    
                if event.type == KEYDOWN:
                    if event.key == K_ESCAPE:   # Esc closes the game
                        onSimulation = False
                        onTitleScreen = True
                    if event.key == K_SPACE:    # Spacebar toggles max range circles
                        rangeEnabled = not rangeEnabled

                    if event.key == K_UP:       # Press UP to increase step size
                        STEP_SIZE = min(STEP_SIZE+.05,1)
                        HCONST = STEP_SIZE/6
                    if event.key == K_DOWN:     # Press DOWN to decrease step size
                        STEP_SIZE = max(STEP_SIZE-.05,0)
                        HCONST = STEP_SIZE/6
                    
                    if event.key == K_a:    # If A is pressed, select all friendly units
                        for unit in Unit.friendlyList:
                            Unit.selectedList.append(unit)
                            unit.isSelected = True
                            unit.color = FRIENDLY_COLOR2
                    if event.key == K_z:    # If Z is pressed, de-select all units
                        Unit.selectedList = []
                        for unit in Unit.friendlyList:
                            unit.isSelected = False
                            unit.color = FRIENDLY_COLOR
                            unit.waypoint(False)
                    if event.key == K_p:
                        paused = not paused
                    if event.key == K_f:
                        fow = not fow
    ### End of simulation loop
    
    if onHelpScreen:    # Help screen
        file = open(HELP_MENU_FILE,"r")
        helpText = [] # Initialize text array
        sectionNum = pageNum = 0
        helpSections = []
        newSection = True
        for line in file:   # Read all lines in the file
            line = str(line.split("\n")[0])
            if newSection:
                helpSections.append(line.split(":")[0])
                helpText.append([[]])
                newSection = False
            if line == "-": # If a line contains only -, new page
                pageNum += 1
                helpText[sectionNum].append(list())
            elif line == "---": # If a line contains only ---, new section
                sectionNum += 1
                pageNum = 0
                newSection = True
            else:
                helpText[sectionNum][pageNum].append(line)
                
        sectionID = 0
        page = 0
        helpScreenUpdate()
        while onHelpScreen:
            titleOffset -= TITLE_SCROLL_SPEED # Moving background
            helpScreenUpdate()
            for event in pygame.event.get():        
                if event.type == QUIT:
                    os._exit(99)
                    
                if event.type == KEYDOWN:
                    if event.key == K_ESCAPE:   # Esc exits to title scren
                        onHelpScreen = False
                        onTitleScreen = True
                    if event.key == K_UP:   # Up/down changes sections
                        page = 0
                        if sectionID > 0: 
                            sectionID -= 1
                        else:
                            sectionID = len(helpText)-1
                    if event.key == K_DOWN:
                        page = 0
                        if sectionID < len(helpText)-1:
                            sectionID += 1
                        else:
                            sectionID = 0
                        
                    if event.key == K_LEFT and page > 0:    # Left/right changes pages
                        page -= 1
                    if event.key == K_RIGHT and page < len(helpText[sectionID])-1:
                        page += 1
