# Romeo and Juliet love simulator program
# Solves a set of two ODEs using RK4 and graphs the output for
# various initial values of R and J.
# By William Ho

print("Romeo and Juliet love simulator program",
      "This program solves a set of two ODEs using RK4 and graphs the output",
      "for various initial values of R and J around the origin.",
      "R (the x-axis) represents Romeo's love for Juliet, while J (the y-axis)",
      "represents Juliet's love for Romeo.\n",
      sep="\n")

import math, pygame
from pygame.locals import *

# Define values for step size, initial and final times
h = 0.1
hConst = h/6        # A constant used in RK4
t = tInitial = 0

# Define variables for graphical output
width = height = 500
bgColor = (255,255,255)
axisColor = (0,0,0)
padding = 0.1   # Amount of empty space at each edge of the screen
                # This value should be much less than 0.5

def dInput(string,default):
    """A version of the input() function with a default
        value if nothing is entered"""
    print(string," [",default,"]",sep="",end=" ")
    inputValue = input()
    if inputValue == "":
        inputValue = default
    else:
        inputValue = inputValue
    return inputValue

# Allow the user to input initial conditions, or use default values
print("Input coefficients a, b, c, and d for the set of ODEs:",
      "(Pressing enter uses the specified default value)\n",
      "\tdR/dt = aR + bJ",
      "\tdJ/dt = cR + dJ\n",
      sep="\n")

readFile = dInput("Read values from local file? (y/n)","n")
if readFile.lower() == "n":
    a = float(dInput("Value of a:", 0))
    b = float(dInput("Value of b:", 1))
    c = float(dInput("Value of c:", -1))
    d = float(dInput("Value of d:", 0))
    tFinal = dInput("End time:", 100)
else:   # If anything other than 'n' is entered, it is taken to mean 'yes'
    fileName = dInput("Name of the file?", "ho_rj_input_june14.txt")
    file = open(fileName,"r")
    a = float(file.readline())
    b = float(file.readline())
    c = float(file.readline())
    d = float(file.readline())
    tFinal = float(file.readline())
    print("Value of a:",a)
    print("Value of b:",b)
    print("Value of c:",c)
    print("Value of d:",d)
    print("Value of tFinal:",tFinal)
    file.close()

def RSlope(R,J):
    """ODE #1, returns the value of dR/dt"""
    return a*R+b*J

def JSlope(R,J):
    """ODE #2, returns the value of dJ/dt"""
    return c*R+d*J

def RK4(h,R,J):
    """Estimates the next value of both ODEs using Runge Kutta 4"""
    m1 = RSlope(R,J)
    k1 = JSlope(R,J)
    m2 = RSlope(R+0.5*h*m1,J+0.5*h*k1)
    k2 = JSlope(R+0.5*h*m1,J+0.5*h*k1)
    m3 = RSlope(R+0.5*h*m2,J+0.5*h*k2)
    k3 = JSlope(R+0.5*h*m2,J+0.5*h*k2)
    m4 = RSlope(R+h*m3,J+h*k3)
    k4 = JSlope(R+h*m3,J+h*k3)
    R += (m1+2*m2+2*m3+m4)*hConst
    J += (k1+2*k2+2*k3+k4)*hConst
    return R,J

# Estimates values using RK4 and determines the max/min values of R and J
maxValue = 0   # Initialize maxValue variable
for RInitial in range(-2,3,1):         # Begin loop to evaluate an array of curves
    for JInitial in range(-2,3,1):     # with different initial R and J values

        # Reset values of R and J at the beginning of each iteration
        R = RInitial
        J = JInitial
        t = tInitial
        
        while t < tFinal+hConst:    # Determine max/min value
            R,J = RK4(h,R,J)
            if abs(R) > maxValue:
                maxValue = abs(R)
            if abs(J) > maxValue:
                maxValue = abs(J)
            t += h

# Determine the scale of the graph based on max/min values
# The padding term allows for some empty space at the edges of the window
scale = width/(2*maxValue)*(1-padding)

# Load a screen for graphical output
pygame.init()
screen = pygame.display.set_mode((width,height))
pygame.display.set_caption('R&J Love Simulation')
screen.fill(bgColor)

# Draw x and y axes on the screen
xCenter = width/2
yCenter = height/2
pygame.draw.line(screen,axisColor,(0,yCenter),(width,yCenter),1)
pygame.draw.line(screen,axisColor,(xCenter,0),(xCenter,height),1)
pygame.display.flip()

def xCoord(R):
    """Converts the value of R into an x value on the graph"""
    return scale*R+xCenter

def yCoord(J):
    """Converts the value of J into a y value on the graph"""
    return scale*J+yCenter

for RInitial in range(-2,3,1):         # Begin loop to draw an array of curves
    for JInitial in range(-2,3,1):     # with different initial R and J values
        lineColor = ((RInitial+3)*(255/5),(JInitial+3)*(255/5),128)   # Diff colors for curves

        # Reset values of R and J
        R = RInitial
        J = JInitial
        t = tInitial
        # EInitial = .5*RSlope(R,J)**2+.5*JSlope(R,J)**2

        # Draw the solution curves on the screen
        while t < tFinal+hConst:
            RNew,JNew = RK4(h,R,J)
            pygame.draw.aaline(screen,lineColor,(xCoord(R),yCoord(J)),(xCoord(RNew),yCoord(JNew)))
            if t == tInitial:       # Draw a colored circle at the initial point on the curve
                pygame.draw.circle(screen,lineColor,(math.floor(xCoord(R)),math.floor(yCoord(J))),3)
            R,J = RNew,JNew         # Update the value of RNew and JNew
            t += h

        # Draw a colored square at the final point on the curve
        pygame.draw.rect(screen, lineColor, (xCoord(R)-3,yCoord(J)-3,6,6), 0)
        pygame.display.flip()   # Refresh screen

        # Error checking: used for debugging
        # EFinal = .5*RSlope(R,J)**2+.5*JSlope(R,J)**2
        # if EInitial != 0:
        #    print("Error at (",RInitial,",",JInitial,"): ",(EFinal - EInitial)/EInitial*100,sep="")

# The following allows the user to close the screen before plotting has completed
running = True

while running:
    for event in pygame.event.get():
        # When you press the 'X' on the top right corner, the program closes
        if event.type == QUIT:
            running = False
        # When you press 'ESC' on the keyboard, the program closes
        if event.type == KEYDOWN and event.key == K_ESCAPE:
            running = False

pygame.display.quit()
